"""Test that the recipes are updated correctly."""

from pathlib import Path

import pandas as pd
import pytest
from climate_ref_esmvaltool import provider
from climate_ref_esmvaltool.diagnostics.base import ESMValToolDiagnostic
from pytest_regressions.file_regression import FileRegressionFixture

from climate_ref.solver import solve_executions
from climate_ref_core.datasets import SourceDatasetType


@pytest.mark.parametrize(
    "diagnostic", [pytest.param(diagnostic, id=diagnostic.slug) for diagnostic in provider.diagnostics()]
)
def test_write_recipe(
    tmp_path: Path,
    file_regression: FileRegressionFixture,
    data_catalog: dict[SourceDatasetType, pd.DataFrame],
    diagnostic: ESMValToolDiagnostic,
) -> None:
    """Test that the recipes are updated correctly."""

    def get_source_types(execution):
        return tuple(sorted(k.value for k in execution.datasets.keys()))

    seen = set()
    for execution in solve_executions(
        data_catalog=data_catalog,
        diagnostic=diagnostic,
        provider=diagnostic.provider,
    ):
        if (source_types := get_source_types(execution)) not in seen:
            seen.add(source_types)
            definition = execution.build_execution_definition(output_root=tmp_path)
            definition.output_directory.mkdir(parents=True, exist_ok=True)
            tmp_recipe = diagnostic.write_recipe(definition=definition)
            reference_recipe = (
                Path(__file__).parent
                / "recipes"
                / f"recipe-{diagnostic.slug}-{'-'.join(source_types)}.yml".replace("-", "_")
            )
            encoding = "utf-8"
            file_regression.check(
                tmp_recipe.read_text(encoding),
                encoding=encoding,
                fullpath=reference_recipe,
            )



from pathlib import Path

import pandas as pd

from climate_ref_core.constraints import (
    AddSupplementaryDataset,
    PartialDateTime,
    RequireFacets,
    RequireTimerange,
)
from climate_ref_core.datasets import ExecutionDatasetCollection, FacetFilter, SourceDatasetType
from climate_ref_core.diagnostics import DataRequirement
from climate_ref_core.esgf import CMIP6Request
from climate_ref_core.metric_values.typing import FileDefinition
from climate_ref_core.pycmec.metric import CMECMetric, MetricCV
from climate_ref_core.pycmec.output import CMECOutput
from climate_ref_core.testing import TestCase, TestDataSpecification
from climate_ref_esmvaltool.diagnostics.base import ESMValToolDiagnostic, get_cmip_source_type
from climate_ref_esmvaltool.recipe import dataframe_to_recipe
from climate_ref_esmvaltool.types import MetricBundleArgs, OutputBundleArgs, Recipe


class ClimateAtGlobalWarmingLevels(ESMValToolDiagnostic):
    """
    Calculate climate variables at global warming levels.
    """

    name = "Climate variables at global warming levels"
    slug = "climate-at-global-warming-levels"
    base_recipe = "recipe_calculate_gwl_exceedance_stats.yml"

    variables = (
        "pr",
        "tas",
    )

    matching_facets = (
        "source_id",
        "member_id",
        "grid_label",
        "table_id",
        "variable_id",
    )

    cmip7_matching_facets = (
        "source_id",
        "variant_label",
        "grid_label",
        "variable_id",
    )

    data_requirements = (
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP6,
                filters=(
                    FacetFilter(
                        facets={
                            "variable_id": variables,
                            "experiment_id": (
                                "ssp126",
                                "ssp245",
                                "ssp370",
                                "ssp585",
                            ),
                            "table_id": "Amon",
                        },
                    ),
                ),
                group_by=("experiment_id",),
                constraints=(
                    AddSupplementaryDataset(
                        supplementary_facets={"experiment_id": "historical"},
                        matching_facets=matching_facets,
                        optional_matching_facets=tuple(),
                    ),
                    RequireTimerange(
                        group_by=matching_facets,
                        start=PartialDateTime(year=1850, month=1),
                        end=PartialDateTime(year=2100, month=12),
                    ),
                    RequireFacets(
                        "experiment_id",
                        required_facets=("historical",),
                        group_by=matching_facets,
                    ),
                    RequireFacets(
                        "variable_id",
                        required_facets=variables,
                        group_by=("experiment_id", "source_id", "member_id", "grid_label", "table_id"),
                    ),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP6),
                ),
            ),
        ),
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP7,
                filters=(
                    FacetFilter(
                        facets={
                            "experiment_id": (
                                # TODO: Redetermine the scenario naming for CMIP7 and update these accordingly
                            ),
                            "frequency": "mon",
                            "region": "glb",
                            "realm": "atmos",
                        },
                    ),
                ),
                group_by=("experiment_id",),
                constraints=(
                    AddSupplementaryDataset(
                        supplementary_facets={"experiment_id": "historical"},
                        matching_facets=cmip7_matching_facets,
                        optional_matching_facets=tuple(),
                    ),
                    RequireTimerange(
                        group_by=cmip7_matching_facets,
                        start=PartialDateTime(year=1850, month=1),
                        end=PartialDateTime(year=2100, month=12),
                    ),
                    RequireFacets(
                        "experiment_id",
                        required_facets=("historical",),
                        group_by=cmip7_matching_facets,
                    ),
                    RequireFacets(
                        "variable_id",
                        required_facets=variables,
                        group_by=("experiment_id", "source_id", "variant_label", "grid_label"),
                    ),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP7),
                ),
            ),
        ),
    )
    facets = ("experiment_id", "global warming level", "metric")

    files = tuple(
        FileDefinition(
            file_pattern=f"plots/gwl_mean_plots_{var_name}/plot_gwl_stats/*.png",
            dimensions={
                "statistic": "mean",
                "variable_id": var_name,
            },
        )
        for var_name in variables
    ) + tuple(
        FileDefinition(
            file_pattern=f"work/gwl_mean_plots_{var_name}/plot_gwl_stats/*.nc",
            dimensions={
                "statistic": "mean",
                "variable_id": var_name,
            },
        )
        for var_name in variables
    )

    test_data_spec = TestDataSpecification(
        test_cases=(
            TestCase(
                name="cmip6",
                description="Test with CMIP6 data.",
                requests=(
                    CMIP6Request(
                        slug="cmip6",
                        facets={
                            "experiment_id": ["ssp245", "historical"],
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", "pr", "tas"],
                            "frequency": ["fx", "mon"],
                        },
                        remove_ensembles=True,
                        time_span=("1850", "2100"),
                    ),
                ),
            ),
            # Disabling test until we have scenarios available for CMIP7
            # TestCase(
            #     name="cmip7",
            #     description="Test with CMIP7 data.",
            #     requests=(
            #         CMIP7Request(
            #             slug="cmip7",
            #             facets={
            #                 "experiment_id": ["ssp245", "historical"],
            #                 "source_id": "CanESM5",
            #                 "variable_id": ["areacella", "pr", "tas"],
            #                 "branded_variable": [
            #                     "areacella_ti-u-hxy-u",
            #                     "pr_tavg-u-hxy-u",
            #                     "tas_tavg-h2m-hxy-u",
            #                 ],
            #                 "variant_label": "r1i1p1f1",
            #                 "frequency": ["fx", "mon"],
            #                 "region": "glb",
            #             },
            #             remove_ensembles=True,
            #             time_span=("1850", "2100"),
            #         ),
            #     ),
            # ),
        )
    )

    @staticmethod
    def update_recipe(
        recipe: Recipe,
        input_files: dict[SourceDatasetType, pd.DataFrame],
    ) -> None:
        """Update the recipe."""
        # Set up the datasets
        cmip_source = get_cmip_source_type(input_files)
        diagnostics = recipe["diagnostics"]
        for diagnostic in diagnostics.values():
            diagnostic.pop("additional_datasets")
        group_by: tuple[str, ...]
        if cmip_source == SourceDatasetType.CMIP7:
            group_by = (
                "source_id",
                "variant_label",
                "grid_label",
                "variable_id",
            )
        else:
            group_by = (
                "source_id",
                "member_id",
                "grid_label",
                "table_id",
                "variable_id",
            )
        recipe_variables = dataframe_to_recipe(
            input_files[cmip_source],
            group_by=group_by,
        )
        datasets = recipe_variables["tas"]["additional_datasets"]
        datasets = [ds for ds in datasets if ds["exp"] != "historical"]
        for dataset in datasets:
            dataset.pop("timerange")
        recipe["datasets"] = datasets

        # Specify the timeranges
        diagnostics["calculate_gwl_exceedance_years"]["variables"]["tas_anomaly"] = {
            "short_name": "tas",
            "preprocessor": "calculate_anomalies",
            "timerange": "1850/2100",
        }

        diagnostics["gwl_mean_plots_tas"]["variables"]["tas"] = {
            "short_name": "tas",
            "preprocessor": "multi_model_gwl_stats",
            "timerange": "2000/2100",
        }

        diagnostics["gwl_mean_plots_pr"]["variables"]["pr"] = {
            "short_name": "pr",
            "preprocessor": "multi_model_gwl_stats",
            "timerange": "2000/2100",
        }

    @staticmethod
    def format_result(
        result_dir: Path,
        execution_dataset: ExecutionDatasetCollection,
        metric_args: MetricBundleArgs,
        output_args: OutputBundleArgs,
    ) -> tuple[CMECMetric, CMECOutput]:
        """Format the result."""
        metric_args[MetricCV.DIMENSIONS.value] = {
            "json_structure": [
                "global warming level",
                "metric",
            ],
            "global warming level": {},
            "metric": {"exceedance_year": {}},
        }

        df = pd.read_csv(
            result_dir
            / "work"
            / "calculate_gwl_exceedance_years"
            / "gwl_exceedance_calculation"
            / "GWL_exceedance_years.csv"
        )
        for row in df.itertuples(index=False):
            gwl = str(row.GWL)
            if gwl not in metric_args[MetricCV.DIMENSIONS.value]["global warming level"]:
                metric_args[MetricCV.DIMENSIONS.value]["global warming level"][gwl] = {}
            metric_args[MetricCV.RESULTS.value][gwl] = {
                "exceedance_year": int(str(row.Exceedance_Year)),
            }

        return CMECMetric.model_validate(metric_args), CMECOutput.model_validate(output_args)



"""
Rapid evaluating CMIP data with ESMValTool.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pooch
from loguru import logger

import climate_ref_esmvaltool.diagnostics
from climate_ref_core.dataset_registry import (
    DATASET_URL,
    dataset_registry_manager,
    fetch_all_files,
    validate_registry_cache,
)
from climate_ref_core.providers import CondaDiagnosticProvider
from climate_ref_esmvaltool._version import __version__
from climate_ref_esmvaltool.recipe import _ESMVALCORE_URL, _ESMVALTOOL_URL

if TYPE_CHECKING:
    from climate_ref.config import Config

_REGISTRY_NAME = "esmvaltool"


class ESMValToolProvider(CondaDiagnosticProvider):
    """Provider for ESMValTool diagnostics."""

    def fetch_data(self, config: Config) -> None:
        """Fetch ESMValTool reference data."""
        registry = dataset_registry_manager[_REGISTRY_NAME]
        fetch_all_files(registry, _REGISTRY_NAME, output_dir=None)

    def validate_setup(self, config: Config) -> bool:
        """Validate conda environment and data checksums."""
        # First check conda environment
        if not super().validate_setup(config):
            return False

        # Then check data checksums
        registry = dataset_registry_manager[_REGISTRY_NAME]
        errors = validate_registry_cache(registry, _REGISTRY_NAME)
        if errors:
            for error in errors:
                logger.error(f"{self.slug} validation failed: {error}")
            logger.error(
                f"Data for {self.slug} is missing or corrupted. "
                f"Please run `ref providers setup --provider {self.slug}` to fetch data."
            )
            return False
        return True

    def get_data_path(self) -> Path | None:
        """Get the path where ESMValTool data is cached."""
        return Path(pooch.os_cache("climate_ref"))


# Initialise the diagnostics manager.
provider = ESMValToolProvider(
    "ESMValTool",
    __version__,
)
provider.pip_packages = [_ESMVALTOOL_URL, _ESMVALCORE_URL]

# Register the diagnostics.
for _diagnostic_cls_name in climate_ref_esmvaltool.diagnostics.__all__:
    _diagnostic_cls = getattr(climate_ref_esmvaltool.diagnostics, _diagnostic_cls_name)
    provider.register(_diagnostic_cls())

# Register OBS, OBS6, and raw data
dataset_registry_manager.register(
    "esmvaltool",
    base_url=DATASET_URL,
    package="climate_ref_esmvaltool.dataset_registry",
    resource="data.txt",
)



from pathlib import Path

import pandas
import xarray

from climate_ref_core.constraints import (
    AddParentDataset,
    AddSupplementaryDataset,
    RequireContiguousTimerange,
    RequireFacets,
)
from climate_ref_core.datasets import ExecutionDatasetCollection, FacetFilter, SourceDatasetType
from climate_ref_core.diagnostics import DataRequirement
from climate_ref_core.esgf import CMIP6Request, CMIP7Request
from climate_ref_core.metric_values.typing import FileDefinition
from climate_ref_core.pycmec.metric import CMECMetric, MetricCV
from climate_ref_core.pycmec.output import CMECOutput
from climate_ref_core.testing import TestCase, TestDataSpecification
from climate_ref_esmvaltool.diagnostics.base import (
    ESMValToolDiagnostic,
    fillvalues_to_nan,
    get_cmip_source_type,
)
from climate_ref_esmvaltool.recipe import dataframe_to_recipe, get_child_and_parent_dataset
from climate_ref_esmvaltool.types import MetricBundleArgs, OutputBundleArgs, Recipe


class TransientClimateResponseEmissions(ESMValToolDiagnostic):
    """
    Calculate the global mean Transient Climate Response to Cumulative CO2 Emissions.
    """

    name = "Transient Climate Response to Cumulative CO2 Emissions"
    slug = "transient-climate-response-emissions"
    base_recipe = "recipe_tcre.yml"

    variables = (
        "tas",
        "fco2antt",
    )

    data_requirements = (
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP6,
                filters=(
                    FacetFilter(
                        facets={
                            "variable_id": "tas",
                            "experiment_id": "esm-1pctCO2",
                            "table_id": "Amon",
                        },
                    ),
                ),
                group_by=("source_id", "member_id", "grid_label"),
                constraints=(
                    AddParentDataset.from_defaults(SourceDatasetType.CMIP6),
                    AddSupplementaryDataset(
                        supplementary_facets={
                            "variable_id": "fco2antt",
                            "experiment_id": "esm-1pctCO2",
                        },
                        matching_facets=(
                            "source_id",
                            "member_id",
                            "table_id",
                            "grid_label",
                        ),
                        optional_matching_facets=("version",),
                    ),
                    RequireContiguousTimerange(group_by=("instance_id",)),
                    RequireFacets("variable_id", ("tas", "fco2antt")),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP6),
                ),
            ),
        ),
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP7,
                filters=(
                    FacetFilter(
                        facets={
                            "branded_variable": (
                                "fco2antt_tavg-u-hxy-u",
                                "tas_tavg-h2m-hxy-u",
                            ),
                            "experiment_id": "esm-1pctCO2",
                            "frequency": "mon",
                            "region": "glb",
                            "realm": "atmos",
                        },
                    ),
                    FacetFilter(
                        facets={
                            "branded_variable": "tas_tavg-h2m-hxy-u",
                            "experiment_id": "esm-piControl",
                            "frequency": "mon",
                            "region": "glb",
                            "realm": "atmos",
                        },
                    ),
                ),
                group_by=("source_id", "variant_label", "grid_label"),
                constraints=(
                    RequireContiguousTimerange(group_by=("instance_id",)),
                    RequireFacets("experiment_id", ("esm-1pctCO2", "esm-piControl")),
                    RequireFacets("variable_id", variables),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP7),
                ),
            ),
        ),
    )
    facets = ("grid_label", "member_id", "variant_label", "source_id", "region", "metric")
    # TODO: the ESMValTool diagnostic script does not save the data for the timeseries.
    series = tuple()
    files = (
        FileDefinition(
            file_pattern="plots/tcre/calculate_tcre/*.png",
            dimensions={"statistic": "tcre"},
        ),
        FileDefinition(
            file_pattern="work/tcre/calculate_tcre/tcre.nc",
            dimensions={"metric": "tcre"},
        ),
    )

    test_data_spec = TestDataSpecification(
        test_cases=(
            TestCase(
                name="cmip6",
                description="Test with CMIP6 data.",
                requests=(
                    CMIP6Request(
                        slug="cmip6",
                        facets={
                            "experiment_id": ["esm-1pctCO2", "esm-piControl"],
                            "source_id": "MPI-ESM1-2-LR",
                            "variable_id": ["areacella", "fco2antt", "tas"],
                            "frequency": ["fx", "mon"],
                        },
                        remove_ensembles=True,
                    ),
                ),
            ),
            TestCase(
                name="cmip7",
                description="Test with CMIP7 data.",
                requests=(
                    CMIP7Request(
                        slug="cmip7",
                        facets={
                            "experiment_id": ["esm-1pctCO2", "esm-piControl"],
                            "source_id": "MPI-ESM1-2-LR",
                            "variable_id": ["areacella", "fco2antt", "tas"],
                            "branded_variable": [
                                "areacella_ti-u-hxy-u",
                                "fco2antt_tavg-u-hxy-u",
                                "tas_tavg-h2m-hxy-u",
                            ],
                            "variant_label": "r1i1p1f1",
                            "frequency": ["fx", "mon"],
                            "region": "glb",
                        },
                        remove_ensembles=True,
                    ),
                ),
            ),
        )
    )

    @staticmethod
    def update_recipe(
        recipe: Recipe,
        input_files: dict[SourceDatasetType, pandas.DataFrame],
    ) -> None:
        """Update the recipe."""
        # Prepare updated datasets section in recipe. It contains three
        # datasets, "tas" and "fco2antt" for the "esm-1pctCO2" and just "tas"
        # for the "esm-piControl" experiment.
        cmip_source = get_cmip_source_type(input_files)
        df = input_files[cmip_source]
        tas_esm_1pctCO2, tas_esm_piControl = get_child_and_parent_dataset(
            df[df.variable_id == "tas"],
            parent_experiment="esm-piControl",
            child_duration_in_years=65,
            parent_offset_in_years=0,
            parent_duration_in_years=65,
        )
        recipe_variables = dataframe_to_recipe(df[df.variable_id == "fco2antt"])

        fco2antt_esm_1pctCO2 = next(
            ds for ds in recipe_variables["fco2antt"]["additional_datasets"] if ds["exp"] == "esm-1pctCO2"
        )
        fco2antt_esm_1pctCO2["timerange"] = tas_esm_1pctCO2["timerange"]

        recipe["diagnostics"]["tcre"]["variables"] = {
            "tas_esm-1pctCO2": {
                "short_name": "tas",
                "preprocessor": "global_annual_mean_anomaly",
                "additional_datasets": [tas_esm_1pctCO2],
            },
            "tas_esm-piControl": {
                "short_name": "tas",
                "preprocessor": "global_annual_mean_anomaly",
                "additional_datasets": [tas_esm_piControl],
            },
            "fco2antt": {
                "preprocessor": "global_cumulative_sum",
                "additional_datasets": [fco2antt_esm_1pctCO2],
            },
        }
        recipe["diagnostics"].pop("barplot")

        # Update descriptions.
        dataset = tas_esm_1pctCO2["dataset"]
        ensemble = tas_esm_1pctCO2["ensemble"]
        settings = recipe["diagnostics"]["tcre"]["scripts"]["calculate_tcre"]
        settings["caption"] = (
            settings["caption"].replace("MPI-ESM1-2-LR", dataset).replace("r1i1p1f1", ensemble)
        )
        settings["pyplot_kwargs"]["title"] = (
            settings["pyplot_kwargs"]["title"].replace("MPI-ESM1-2-LR", dataset).replace("r1i1p1f1", ensemble)
        )

    @staticmethod
    def format_result(
        result_dir: Path,
        execution_dataset: ExecutionDatasetCollection,
        metric_args: MetricBundleArgs,
        output_args: OutputBundleArgs,
    ) -> tuple[CMECMetric, CMECOutput]:
        """Format the result."""
        tcre_ds = xarray.open_dataset(result_dir / "work" / "tcre" / "calculate_tcre" / "tcre.nc")
        tcre = float(fillvalues_to_nan(tcre_ds["tcre"].values)[0])

        # Update the diagnostic bundle arguments with the computed diagnostics.
        metric_args[MetricCV.DIMENSIONS.value] = {
            "json_structure": [
                "region",
                "metric",
            ],
            "region": {"global": {}},
            "metric": {"tcre": {}},
        }
        metric_args[MetricCV.RESULTS.value] = {
            "global": {
                "tcre": tcre,
            },
        }
        return CMECMetric.model_validate(metric_args), CMECOutput.model_validate(output_args)



import json

import climate_ref_esmvaltool.diagnostics.base
import numpy as np
import pandas
import pytest
import xarray as xr
import yaml
from climate_ref_esmvaltool.diagnostics.base import ESMValToolDiagnostic
from climate_ref_esmvaltool.types import Recipe

from climate_ref_core.datasets import SourceDatasetType
from climate_ref_core.metric_values import SeriesMetricValue as SeriesMetricValueType
from climate_ref_core.metric_values.typing import SeriesDefinition
from climate_ref_core.pycmec.controlled_vocabulary import CV
from climate_ref_core.pycmec.output import OutputCV


@pytest.fixture
def mock_diagnostic():
    class MockDiagnostic(ESMValToolDiagnostic):
        base_recipe = "examples/recipe_python.yml"

        def update_recipe(
            self,
            recipe: Recipe,
            input_files: dict[SourceDatasetType, pandas.DataFrame],
        ) -> None:
            pass

    return MockDiagnostic()


@pytest.mark.parametrize("data_dir_exists", [True, False])
def test_build_cmd(mocker, tmp_path, metric_definition, mock_diagnostic, data_dir_exists):
    dataset_registry_manager = mocker.patch.object(
        climate_ref_esmvaltool.diagnostics.base,
        "dataset_registry_manager",
    )
    data_dir = tmp_path / "ESMValTool"
    if data_dir_exists:
        data_dir.mkdir()
    dataset_registry_manager.__getitem__.return_value.abspath = tmp_path
    output_dir = metric_definition.output_directory
    output_dir.mkdir(parents=True)
    cmd = mock_diagnostic.build_cmd(metric_definition)
    config_dir = output_dir / "config"
    recipe = output_dir / "recipe.yml"
    assert cmd == ["esmvaltool", "run", f"--config-dir={config_dir}", f"{recipe}"]
    assert (output_dir / "climate_data").is_dir()
    config = yaml.safe_load((config_dir / "config.yml").read_text(encoding="utf-8"))
    assert config["search_data"] == "quick"
    if data_dir_exists:
        assert len(config["projects"]) == 6
        assert "OBS" in config["projects"]
        assert "OBS6" in config["projects"]
        assert "native6" in config["projects"]
        # obs4MIPs should have both local and esmvaltool data sources
        assert "esmvaltool" in config["projects"]["obs4MIPs"]["data"]
    else:
        assert len(config["projects"]) == 3
    assert "CMIP6" in config["projects"]
    assert "CMIP7" in config["projects"]
    assert "obs4MIPs" in config["projects"]


def test_build_metric_result(metric_definition, mock_diagnostic):
    results_dir = metric_definition.to_output_path("executions") / "recipe_test"

    for subdir in "timeseries", "map":
        metadata = {}
        for dirname in "work", "plots":
            for i in range(2):
                suffix = ".nc" if dirname == "work" else ".png"
                filepath = results_dir / dirname / subdir / "script1" / f"file{i}{suffix}"
                metadata[str(filepath)] = {
                    "caption": f"This is {subdir} test file {i}.",
                }
        metadata_file = results_dir / "run" / subdir / "script1" / "diagnostic_provenance.yml"
        metadata_file.parent.mkdir(parents=True)
        with metadata_file.open("w", encoding="utf-8") as file:
            yaml.safe_dump(metadata, file)

    execution_result = mock_diagnostic.build_execution_result(definition=metric_definition)
    metric_bundle = json.loads(
        execution_result.to_output_path(execution_result.metric_bundle_filename).read_text(encoding="utf-8")
    )
    output_bundle = json.loads(
        execution_result.to_output_path(execution_result.output_bundle_filename).read_text(encoding="utf-8")
    )

    assert isinstance(metric_bundle, dict)
    assert metric_bundle

    assert isinstance(output_bundle, dict)
    assert OutputCV.DATA.value in output_bundle
    assert len(output_bundle[OutputCV.DATA.value]) == 4
    assert OutputCV.PLOTS.value in output_bundle
    plots = output_bundle[OutputCV.PLOTS.value]
    assert len(plots) == 4
    captions = {p["long_name"] for p in plots.values()}
    assert len(captions) == 4


def test_series_extraction(tmp_path, metric_definition, mock_diagnostic, mocker):
    # Definition of the netcdf files to extract series from
    metric_definition.diagnostic.series = [
        SeriesDefinition(
            file_pattern="work/timeseries/script1/file0.nc",
            attributes=["long_name", "units"],
            dimensions={"model": "TestModel"},
            values_name="data",
            index_name="time",
        )
    ]

    # Create a NetCDF file matching the pattern
    results_dir = metric_definition.to_output_path("executions") / "recipe_test"
    nc_path = results_dir / "work" / "timeseries" / "script1" / "file0.nc"
    nc_path.parent.mkdir(parents=True, exist_ok=True)
    times = np.array([1, 2, 3])
    data = np.array([10.0, 20.0, 30.0])
    ds = xr.Dataset({"data": ("time", data)}, coords={"time": times})
    ds.attrs["long_name"] = "Test Data"
    ds.attrs["units"] = "K"
    ds.to_netcdf(nc_path)

    # Dummy metadata file
    metadata_file = results_dir / "run" / "timeseries" / "script1" / "diagnostic_provenance.yml"
    metadata_file.parent.mkdir(parents=True, exist_ok=True)
    metadata = {str(nc_path): {"caption": "Test caption."}}
    with metadata_file.open("w", encoding="utf-8") as file:
        yaml.dump(metadata, file)

    result = mock_diagnostic.build_execution_result(definition=metric_definition)

    # Load the series from the output file
    assert result.series_filename is not None, "Series filename should be set"
    loaded_series = SeriesMetricValueType.load_from_json(result.to_output_path(result.series_filename))
    assert loaded_series, "Series should not be empty"
    s = loaded_series[0]
    assert isinstance(s, SeriesMetricValueType)
    assert s.dimensions == {"model": "TestModel"}
    assert s.values == [10.0, 20.0, 30.0]
    assert s.index == [1, 2, 3]
    assert s.index_name == "time"
    assert s.attributes["long_name"] == "Test Data"
    assert s.attributes["units"] == "K"
    assert s.attributes["caption"] == "Test caption."


def test_series_validation_failure(tmp_path, metric_definition, mock_diagnostic, mocker):
    metric_definition.diagnostic.series = [
        SeriesDefinition(
            file_pattern="work/timeseries/script1/file0.nc",
            attributes=["long_name", "units"],
            dimensions={"model": "TestModel"},
            values_name="data",
            index_name="time",
        )
    ]
    results_dir = metric_definition.to_output_path("executions") / "recipe_test"
    nc_path = results_dir / "work" / "timeseries" / "script1" / "file0.nc"
    nc_path.parent.mkdir(parents=True, exist_ok=True)
    times = np.array([1, 2, 3])
    data = np.array([10.0, 20.0, 30.0])
    ds = xr.Dataset({"data": ("time", data)}, coords={"time": times})
    ds.attrs["long_name"] = "Test Data"
    ds.attrs["units"] = "K"
    ds.to_netcdf(nc_path)

    # Dummy metadata file
    metadata_file = results_dir / "run" / "timeseries" / "script1" / "diagnostic_provenance.yml"
    metadata_file.parent.mkdir(parents=True, exist_ok=True)
    metadata = {str(nc_path): {"caption": "Test caption."}}
    with metadata_file.open("w", encoding="utf-8") as file:
        yaml.dump(metadata, file)

    # Patch CV.validate_metrics to raise an error for series
    mocker.patch.object(CV, "validate_metrics", side_effect=AssertionError("Validation failed"))
    log_spy = mocker.spy(climate_ref_esmvaltool.diagnostics.base.logger, "exception")

    # Run build_execution_result (should log exception)
    mock_diagnostic.build_execution_result(definition=metric_definition)
    assert log_spy.call_count >= 0  # Should log the validation failure



import copy
from pathlib import Path

import numpy as np
import pandas
import xarray

from climate_ref_core.constraints import (
    AddSupplementaryDataset,
    PartialDateTime,
    RequireTimerange,
)
from climate_ref_core.datasets import ExecutionDatasetCollection, FacetFilter, SourceDatasetType
from climate_ref_core.diagnostics import DataRequirement
from climate_ref_core.esgf import CMIP6Request, CMIP7Request, Obs4MIPsRequest
from climate_ref_core.metric_values.typing import FileDefinition, SeriesDefinition
from climate_ref_core.pycmec.metric import CMECMetric, MetricCV
from climate_ref_core.pycmec.output import CMECOutput
from climate_ref_core.testing import TestCase, TestDataSpecification
from climate_ref_esmvaltool.diagnostics.base import ESMValToolDiagnostic, fillvalues_to_nan
from climate_ref_esmvaltool.recipe import dataframe_to_recipe
from climate_ref_esmvaltool.types import MetricBundleArgs, OutputBundleArgs, Recipe

REGIONS = (
    "Arabian-Peninsula",
    "Arabian-Sea",
    "Arctic-Ocean",
    "Bay-of-Bengal",
    "C.Australia",
    "C.North-America",
    "Caribbean",
    "Central-Africa",
    "E.Antarctica",
    "E.Asia",
    "E.Australia",
    "E.C.Asia",
    "E.Europe",
    "E.North-America",
    "E.Siberia",
    "E.Southern-Africa",
    "Equatorial.Atlantic-Ocean",
    "Equatorial.Indic-Ocean",
    "Equatorial.Pacific-Ocean",
    "Greenland/Iceland",
    "Madagascar",
    "Mediterranean",
    "N.Atlantic-Ocean",
    "N.Australia",
    "N.Central-America",
    "N.E.North-America",
    "N.E.South-America",
    "N.Eastern-Africa",
    "N.Europe",
    "N.Pacific-Ocean",
    "N.South-America",
    "N.W.North-America",
    "N.W.South-America",
    "New-Zealand",
    "Russian-Arctic",
    "Russian-Far-East",
    "S.Asia",
    "S.Atlantic-Ocean",
    "S.Australia",
    "S.Central-America",
    "S.E.Asia",
    "S.E.South-America",
    "S.Eastern-Africa",
    "S.Indic-Ocean",
    "S.Pacific-Ocean",
    "S.South-America",
    "S.W.South-America",
    "Sahara",
    "South-American-Monsoon",
    "Southern-Ocean",
    "Tibetan-Plateau",
    "W.Antarctica",
    "W.C.Asia",
    "W.North-America",
    "W.Siberia",
    "W.Southern-Africa",
    "West&Central-Europe",
    "Western-Africa",
)


class RegionalHistoricalAnnualCycle(ESMValToolDiagnostic):
    """
    Plot regional historical annual cycle of climate variables.
    """

    name = "Regional historical annual cycle of climate variables"
    slug = "regional-historical-annual-cycle"
    base_recipe = "ref/recipe_ref_annual_cycle_region.yml"

    variables = (
        "hus",
        "pr",
        "psl",
        "tas",
        "ua",
    )

    data_requirements = (
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP6,
                filters=(
                    FacetFilter(
                        facets={
                            "variable_id": variables,
                            "experiment_id": "historical",
                            "table_id": "Amon",
                        },
                    ),
                ),
                group_by=("source_id", "member_id", "grid_label"),
                constraints=(
                    RequireTimerange(
                        group_by=("instance_id",),
                        start=PartialDateTime(1980, 1),
                        end=PartialDateTime(2009, 12),
                    ),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP6),
                ),
            ),
        ),
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP7,
                filters=(
                    FacetFilter(
                        facets={
                            "branded_variable": (
                                "hus_tavg-p19-hxy-u",
                                "pr_tavg-u-hxy-u",
                                "psl_tavg-u-hxy-u",
                                "tas_tavg-h2m-hxy-u",
                                "ua_tavg-p19-hxy-air",
                            ),
                            "experiment_id": "historical",
                            "frequency": "mon",
                            "realm": "atmos",
                        },
                    ),
                ),
                group_by=("source_id", "variant_label", "grid_label"),
                constraints=(
                    RequireTimerange(
                        group_by=("instance_id",),
                        start=PartialDateTime(1980, 1),
                        end=PartialDateTime(2009, 12),
                    ),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP7),
                ),
            ),
        ),
        (
            DataRequirement(
                source_type=SourceDatasetType.obs4MIPs,
                filters=(
                    FacetFilter(
                        facets={
                            "variable_id": (
                                "psl",
                                "ua",
                            ),
                            "source_id": "ERA-5",
                            "frequency": "mon",
                        },
                    ),
                ),
                group_by=("source_id",),
                constraints=(
                    RequireTimerange(
                        group_by=("instance_id",),
                        start=PartialDateTime(1980, 1),
                        end=PartialDateTime(2009, 12),
                    ),
                ),
                # TODO: Add obs4MIPs datasets once available and working:
                #
                # obs4MIPs dataset that cannot be ingested (https://github.com/Climate-REF/climate-ref/issues/260):
                # - GPCP-V2.3: pr
                #
                # Not yet available on obs4MIPs:
                # - ERA5: hus
                # - HadCRUT5_ground_5.0.1.0-analysis: tas
            ),
        ),
    )

    test_data_spec = TestDataSpecification(
        test_cases=(
            TestCase(
                name="cmip6",
                description="Test with CMIP6 data.",
                requests=(
                    CMIP6Request(
                        slug="cmip6",
                        facets={
                            "experiment_id": ["historical"],
                            "frequency": ["fx", "mon"],
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", *variables],
                        },
                        remove_ensembles=True,
                        time_span=("1980", "2009"),
                    ),
                ),
            ),
            TestCase(
                name="cmip7",
                description="Test with CMIP7 data.",
                requests=(
                    CMIP7Request(
                        slug="cmip7",
                        facets={
                            "experiment_id": ["historical"],
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", *variables],
                            "branded_variable": [
                                "areacella_ti-u-hxy-u",
                                "hus_tavg-p19-hxy-u",
                                "pr_tavg-u-hxy-u",
                                "psl_tavg-u-hxy-u",
                                "tas_tavg-h2m-hxy-u",
                                "ua_tavg-p19-hxy-air",
                            ],
                            "variant_label": "r1i1p1f1",
                            "frequency": ["fx", "mon"],
                            "region": "glb",
                        },
                        remove_ensembles=True,
                        time_span=("1980", "2009"),
                    ),
                ),
            ),
            TestCase(
                name="obs4mips",
                description="Test with obs4MIPs data.",
                requests=(
                    Obs4MIPsRequest(
                        slug="obs4mips",
                        facets={
                            "project": "obs4MIPs",
                            "source_id": "ERA-5",
                            "variable_id": [
                                "psl",
                                "ua",
                            ],
                        },
                        remove_ensembles=False,
                        time_span=("1980", "2009"),
                    ),
                ),
            ),
        ),
    )

    facets = ()
    files = tuple(
        FileDefinition(
            file_pattern=f"plots/anncyc-{region}/allplots/*_{var_name}_*.png",
            dimensions={
                "region": region,
                "variable_id": var_name,
                "statistic": "mean",
            },
        )
        for var_name in variables
        for region in REGIONS
    )
    series = tuple(
        SeriesDefinition(
            file_pattern=f"anncyc-{region}/allplots/*_{var_name}_*.nc",
            sel={"dim0": 0},  # Select the model and not the observation.
            dimensions=(
                {
                    "region": region,
                    "variable_id": var_name,
                    "statistic": "mean",
                }
            ),
            values_name=var_name,
            index_name="month_number",
            attributes=[],
        )
        for var_name in variables
        for region in REGIONS
    )

    @staticmethod
    def update_recipe(
        recipe: Recipe,
        input_files: dict[SourceDatasetType, pandas.DataFrame],
    ) -> None:
        """Update the recipe."""
        # Extra datasets that are not published as obs4MIPs.
        extra_datasets: dict[str, list[dict[str, str | int]]] = {
            "tas": [
                {
                    "dataset": "HadCRUT5",
                    "project": "OBS",
                    "tier": 2,
                    "type": "ground",
                    "version": "5.0.1.0-analysis",
                },
            ],
            "pr": [
                {
                    "dataset": "GPCP-V2.3",
                    "project": "obs4MIPs",
                },
            ],
            "hus": [
                {
                    "dataset": "ERA5",
                    "project": "native6",
                    "tier": 3,
                    "type": "reanaly",
                    "version": "v1",
                },
            ],
        }

        # Remove the unused regions alias.
        recipe.pop("regions")
        # Update the dataset.
        recipe.pop("datasets")
        recipe_variables = dataframe_to_recipe(next(iter(input_files.values())))
        source_type = next(iter(input_files.keys()))
        project = str(source_type).split(".")[1]
        for diagnostic_name, diagnostic in dict(recipe["diagnostics"]).items():
            for variable_group, variable in dict(diagnostic["variables"]).items():
                short_name = variable_group.split("_")[0]
                datasets = []
                if short_name in recipe_variables:
                    dataset = copy.deepcopy(recipe_variables[short_name]["additional_datasets"][0])
                    dataset.pop("timerange", None)
                    datasets.append(dataset)
                if project == "obs4MIPs" and short_name in extra_datasets:
                    datasets.extend(copy.deepcopy(extra_datasets[short_name]))
                if datasets:
                    variable["additional_datasets"] = datasets
                else:
                    # If no data is available for a diagnostic, skip it.
                    recipe["diagnostics"].pop(diagnostic_name)
                    continue
            for script_settings in diagnostic["scripts"].values():
                if project == "obs4MIPs":
                    label = "{dataset}"
                else:
                    label = "{dataset}_{ensemble}_{grid}"
                script_settings["plot_filename"] = f"{{plot_type}}_{{real_name}}_{label}_{{shape_id}}"
                for plot_settings in script_settings["plots"].values():
                    plot_settings["plot_kwargs"]["default"]["label"] = label


class RegionalHistoricalTimeSeries(RegionalHistoricalAnnualCycle):
    """
    Plot regional historical mean and anomaly of climate variables.
    """

    name = "Regional historical mean and anomaly of climate variables"
    slug = "regional-historical-timeseries"
    base_recipe = "ref/recipe_ref_timeseries_region.yml"

    variables = (
        "hus",
        "pr",
        "psl",
        "tas",
        "ua",
    )

    data_requirements = (
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP6,
                filters=(
                    FacetFilter(
                        facets={
                            "variable_id": variables,
                            "experiment_id": "historical",
                            "table_id": "Amon",
                        },
                    ),
                ),
                group_by=("source_id", "member_id", "grid_label"),
                constraints=(
                    RequireTimerange(
                        group_by=("instance_id",),
                        start=PartialDateTime(1980, 1),
                        end=PartialDateTime(2014, 12),
                    ),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP6),
                ),
            ),
        ),
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP7,
                filters=(
                    FacetFilter(
                        facets={
                            "branded_variable": (
                                "hus_tavg-p19-hxy-u",
                                "pr_tavg-u-hxy-u",
                                "psl_tavg-u-hxy-u",
                                "tas_tavg-h2m-hxy-u",
                                "ua_tavg-p19-hxy-air",
                            ),
                            "experiment_id": "historical",
                            "frequency": "mon",
                            "region": "glb",
                            "realm": "atmos",
                        },
                    ),
                ),
                group_by=("source_id", "variant_label", "grid_label"),
                constraints=(
                    RequireTimerange(
                        group_by=("instance_id",),
                        start=PartialDateTime(1980, 1),
                        end=PartialDateTime(2014, 12),
                    ),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP7),
                ),
            ),
        ),
        (
            DataRequirement(
                source_type=SourceDatasetType.obs4MIPs,
                filters=(
                    FacetFilter(
                        facets={
                            "variable_id": (
                                "psl",
                                "ua",
                            ),
                            "source_id": "ERA-5",
                            "frequency": "mon",
                        },
                    ),
                ),
                group_by=("source_id",),
                constraints=(
                    RequireTimerange(
                        group_by=("instance_id",),
                        start=PartialDateTime(1980, 1),
                        end=PartialDateTime(2014, 12),
                    ),
                ),
                # TODO: Add obs4MIPs datasets once available and working:
                #
                # obs4MIPs dataset that cannot be ingested (https://github.com/Climate-REF/climate-ref/issues/260):
                # - GPCP-V2.3: pr
                #
                # Not yet available on obs4MIPs:
                # - ERA5: hus
                # - HadCRUT5_ground_5.0.1.0-analysis: tas
            ),
        ),
    )

    test_data_spec = TestDataSpecification(
        test_cases=(
            TestCase(
                name="cmip6",
                description="Test with CMIP6 data.",
                requests=(
                    CMIP6Request(
                        slug="cmip6",
                        facets={
                            "experiment_id": ["historical"],
                            "frequency": ["fx", "mon"],
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", *variables],
                        },
                        remove_ensembles=True,
                        time_span=("1980", "2014"),
                    ),
                ),
            ),
            TestCase(
                name="cmip7",
                description="Test with CMIP7 data.",
                requests=(
                    CMIP7Request(
                        slug="cmip7",
                        facets={
                            "experiment_id": ["historical"],
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", *variables],
                            "branded_variable": [
                                "areacella_ti-u-hxy-u",
                                "hus_tavg-p19-hxy-u",
                                "pr_tavg-u-hxy-u",
                                "psl_tavg-u-hxy-u",
                                "tas_tavg-h2m-hxy-u",
                                "ua_tavg-p19-hxy-air",
                            ],
                            "variant_label": "r1i1p1f1",
                            "frequency": ["fx", "mon"],
                            "region": "glb",
                        },
                        remove_ensembles=True,
                        time_span=("1980", "2014"),
                    ),
                ),
            ),
            TestCase(
                name="obs4mips",
                description="Test with obs4MIPs data.",
                requests=(
                    Obs4MIPsRequest(
                        slug="obs4mips",
                        facets={
                            "project": "obs4MIPs",
                            "source_id": "ERA-5",
                            "variable_id": [
                                "psl",
                                "ua",
                            ],
                        },
                        remove_ensembles=False,
                        time_span=("1980", "2014"),
                    ),
                ),
            ),
        ),
    )

    files = tuple(
        FileDefinition(
            file_pattern=f"plots/{diagnostic}-{region}/allplots/*_{var_name}_*.png",
            dimensions={
                "region": region,
                "variable_id": var_name,
                "statistic": ("mean" if diagnostic == "timeseries_abs" else "mean anomaly"),
            },
        )
        for var_name in variables
        for region in REGIONS
        for diagnostic in ["timeseries_abs", "timeseries"]
    )
    series = tuple(
        SeriesDefinition(
            file_pattern=f"{diagnostic}-{region}/allplots/*_{var_name}_*.nc",
            sel={"dim0": 0},
            dimensions=(
                {
                    "region": region,
                    "variable_id": var_name,
                    "statistic": ("mean" if diagnostic == "timeseries_abs" else "mean anomaly"),
                }
            ),
            values_name=var_name,
            index_name="time",
            attributes=[],
        )
        for var_name in variables
        for region in REGIONS
        for diagnostic in ["timeseries_abs", "timeseries"]
    )


class RegionalHistoricalTrend(ESMValToolDiagnostic):
    """
    Plot regional historical trend of climate variables.
    """

    name = "Regional historical trend of climate variables"
    slug = "regional-historical-trend"
    base_recipe = "ref/recipe_ref_trend_regions.yml"

    variables = (
        "hus",
        "pr",
        "psl",
        "tas",
        "ua",
    )

    data_requirements = (
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP6,
                filters=(
                    FacetFilter(
                        facets={
                            "variable_id": variables,
                            "experiment_id": "historical",
                            "table_id": "Amon",
                        },
                    ),
                ),
                group_by=("source_id", "member_id", "grid_label"),
                constraints=(
                    RequireTimerange(
                        group_by=("instance_id",),
                        start=PartialDateTime(1980, 1),
                        end=PartialDateTime(2009, 12),
                    ),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP6),
                ),
            ),
        ),
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP7,
                filters=(
                    FacetFilter(
                        facets={
                            "branded_variable": (
                                "hus_tavg-p19-hxy-u",
                                "pr_tavg-u-hxy-u",
                                "psl_tavg-u-hxy-u",
                                "tas_tavg-h2m-hxy-u",
                                "ua_tavg-p19-hxy-air",
                            ),
                            "experiment_id": "historical",
                            "frequency": "mon",
                            "realm": "atmos",
                        },
                    ),
                ),
                group_by=("source_id", "variant_label", "grid_label"),
                constraints=(
                    RequireTimerange(
                        group_by=("instance_id",),
                        start=PartialDateTime(1980, 1),
                        end=PartialDateTime(2009, 12),
                    ),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP7),
                ),
            ),
        ),
        (
            DataRequirement(
                source_type=SourceDatasetType.obs4MIPs,
                filters=(
                    FacetFilter(
                        facets={
                            "variable_id": (
                                "psl",
                                "tas",
                                "ua",
                            ),
                            "source_id": "ERA-5",
                            "frequency": "mon",
                        },
                    ),
                ),
                group_by=("source_id",),
                constraints=(
                    RequireTimerange(
                        group_by=("instance_id",),
                        start=PartialDateTime(1980, 1),
                        end=PartialDateTime(2009, 12),
                    ),
                ),
                # TODO: Add obs4MIPs datasets once available and working:
                #
                # obs4MIPs dataset that cannot be ingested (https://github.com/Climate-REF/climate-ref/issues/260):
                # - GPCP-V2.3: pr
                #
                # Not yet available on obs4MIPs:
                # - ERA5: hus, pr
                # - HadCRUT5_ground_5.0.1.0-analysis: tas
            ),
        ),
    )

    test_data_spec = TestDataSpecification(
        test_cases=(
            TestCase(
                name="cmip6",
                description="Test with CMIP6 data.",
                requests=(
                    CMIP6Request(
                        slug="cmip6",
                        facets={
                            "experiment_id": ["historical"],
                            "frequency": ["fx", "mon"],
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", *variables],
                        },
                        remove_ensembles=True,
                        time_span=("1980", "2009"),
                    ),
                ),
            ),
            TestCase(
                name="cmip7",
                description="Test with CMIP7 data.",
                requests=(
                    CMIP7Request(
                        slug="cmip7",
                        facets={
                            "experiment_id": ["historical"],
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", *variables],
                            "branded_variable": [
                                "areacella_ti-u-hxy-u",
                                "hus_tavg-p19-hxy-u",
                                "pr_tavg-u-hxy-u",
                                "psl_tavg-u-hxy-u",
                                "tas_tavg-h2m-hxy-u",
                                "ua_tavg-p19-hxy-air",
                            ],
                            "variant_label": "r1i1p1f1",
                            "frequency": ["fx", "mon"],
                            "region": "glb",
                        },
                        remove_ensembles=True,
                        time_span=("1980", "2009"),
                    ),
                ),
            ),
            TestCase(
                name="obs4mips",
                description="Test with obs4MIPs data.",
                requests=(
                    Obs4MIPsRequest(
                        slug="obs4mips",
                        facets={
                            "project": "obs4MIPs",
                            "source_id": "ERA-5",
                            "variable_id": [
                                "psl",
                                "tas",
                                "ua",
                            ],
                        },
                        remove_ensembles=False,
                        time_span=("1980", "2009"),
                    ),
                ),
            ),
        ),
    )

    facets = ("grid_label", "member_id", "variant_label", "source_id", "variable_id", "region", "metric")
    files = tuple(
        FileDefinition(
            file_pattern=f"plots/{var_name}_trends/plot/seaborn_barplot.png",
            dimensions={
                "variable_id": var_name,
                "statistic": "trend",
            },
        )
        for var_name in ("hus200", "pr", "psl", "tas", "ua200")
    ) + tuple(
        FileDefinition(
            file_pattern=f"work/{var_name}_trends/plot/seaborn_barplot.nc",
            dimensions={
                "variable_id": var_name,
                "statistic": "trend",
            },
        )
        for var_name in ("hus200", "pr", "psl", "tas", "ua200")
    )

    @staticmethod
    def update_recipe(
        recipe: Recipe,
        input_files: dict[SourceDatasetType, pandas.DataFrame],
    ) -> None:
        """Update the recipe."""
        # Extra datasets that are not published as CMIP6-style obs4MIPs.
        extra_datasets: dict[str, list[dict[str, str | int]]] = {
            "tas": [
                {
                    "dataset": "HadCRUT5",
                    "project": "OBS",
                    "tier": 2,
                    "type": "ground",
                    "version": "5.0.1.0-analysis",
                },
            ],
            "pr": [
                {
                    "dataset": "GPCP-V2.3",
                    "project": "obs4MIPs",
                },
                {
                    "dataset": "ERA5",
                    "project": "native6",
                    "type": "reanaly",
                    "tier": 3,
                    "version": "v1",
                },
            ],
            "hus": [
                {
                    "dataset": "ERA5",
                    "project": "native6",
                    "tier": 3,
                    "type": "reanaly",
                    "version": "v1",
                },
            ],
        }

        # Update the datasets.
        recipe.pop("datasets")
        recipe_variables = dataframe_to_recipe(next(iter(input_files.values())))
        source_type = next(iter(input_files.keys()))
        project = str(source_type).split(".")[1]
        for diagnostic_name, diagnostic in dict(recipe["diagnostics"]).items():
            for short_name, variable in dict(diagnostic["variables"]).items():
                datasets = []
                if short_name in recipe_variables:
                    dataset = recipe_variables[short_name]["additional_datasets"][0]
                    dataset.pop("timerange", None)
                    datasets.append(dataset)
                if project == "obs4MIPs" and short_name in extra_datasets:
                    datasets.extend(extra_datasets[short_name])
                if datasets:
                    variable["additional_datasets"] = datasets
                else:
                    # If no data is available for a diagnostic, skip it.
                    recipe["diagnostics"].pop(diagnostic_name)

    @staticmethod
    def format_result(
        result_dir: Path,
        execution_dataset: ExecutionDatasetCollection,
        metric_args: MetricBundleArgs,
        output_args: OutputBundleArgs,
    ) -> tuple[CMECMetric, CMECOutput]:
        """Format the result."""
        metric_args[MetricCV.DIMENSIONS.value] = {
            "json_structure": ["variable_id", "region", "metric"],
            "variable_id": {},
            "region": {},
            "metric": {"trend": {}},
        }
        for file in result_dir.glob("work/*_trends/plot/seaborn_barplot.nc"):
            ds = xarray.open_dataset(file)
            dataset_collection = next(iter(execution_dataset.values()))
            for source_id in dataset_collection.source_id.unique():
                select = source_id == np.array([s.strip() for s in ds.dataset.values.astype(str).tolist()])
                ds.isel(dim0=select)
                variable_id = next(iter(ds.data_vars.keys()))
                metric_args[MetricCV.DIMENSIONS.value]["variable_id"][variable_id] = {}
                metric_args[MetricCV.RESULTS.value][variable_id] = {}
                for region_value, trend_value in zip(
                    ds.shape_id.astype(str).values, fillvalues_to_nan(ds[variable_id].values)
                ):
                    region = region_value.strip()
                    trend = float(trend_value)
                    if region not in metric_args[MetricCV.DIMENSIONS.value]["region"]:
                        metric_args[MetricCV.DIMENSIONS.value]["region"][region] = {}
                    metric_args[MetricCV.RESULTS.value][variable_id][region] = {"trend": trend}

        return CMECMetric.model_validate(metric_args), CMECOutput.model_validate(output_args)



from pathlib import Path

import pandas
import xarray

from climate_ref_core.constraints import (
    AddParentDataset,
    AddSupplementaryDataset,
    RequireContiguousTimerange,
    RequireFacets,
    RequireOverlappingTimerange,
)
from climate_ref_core.datasets import ExecutionDatasetCollection, FacetFilter, SourceDatasetType
from climate_ref_core.diagnostics import DataRequirement
from climate_ref_core.esgf import CMIP6Request, CMIP7Request
from climate_ref_core.metric_values.typing import FileDefinition, SeriesDefinition
from climate_ref_core.pycmec.metric import CMECMetric, MetricCV
from climate_ref_core.pycmec.output import CMECOutput
from climate_ref_core.testing import TestCase, TestDataSpecification
from climate_ref_esmvaltool.diagnostics.base import (
    ESMValToolDiagnostic,
    fillvalues_to_nan,
    get_cmip_source_type,
)
from climate_ref_esmvaltool.recipe import dataframe_to_recipe, get_child_and_parent_dataset
from climate_ref_esmvaltool.types import MetricBundleArgs, OutputBundleArgs, Recipe


class EquilibriumClimateSensitivity(ESMValToolDiagnostic):
    """
    Calculate the global mean equilibrium climate sensitivity for a dataset.
    """

    name = "Equilibrium Climate Sensitivity"
    slug = "equilibrium-climate-sensitivity"
    base_recipe = "recipe_ecs.yml"

    variables = (
        "rlut",
        "rsdt",
        "rsut",
        "tas",
    )
    experiments = (
        "abrupt-4xCO2",
        "piControl",
    )

    data_requirements = (
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP6,
                filters=(
                    FacetFilter(
                        facets={
                            "variable_id": variables,
                            "experiment_id": "abrupt-4xCO2",
                            "table_id": "Amon",
                        },
                    ),
                ),
                group_by=("source_id", "member_id", "grid_label"),
                constraints=(
                    RequireOverlappingTimerange(group_by=("instance_id",)),
                    AddParentDataset.from_defaults(SourceDatasetType.CMIP6),
                    RequireContiguousTimerange(group_by=("instance_id",)),
                    RequireFacets(
                        "variable_id",
                        required_facets=variables,
                        group_by=("source_id", "member_id", "grid_label", "experiment_id"),
                    ),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP6),
                ),
            ),
        ),
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP7,
                filters=(
                    FacetFilter(
                        facets={
                            "branded_variable": (
                                "rlut_tavg-u-hxy-u",
                                "rsdt_tavg-u-hxy-u",
                                "rsut_tavg-u-hxy-u",
                                "tas_tavg-h2m-hxy-u",
                            ),
                            "experiment_id": "abrupt-4xCO2",
                            "frequency": "mon",
                            "region": "glb",
                            "realm": "atmos",
                        },
                    ),
                ),
                group_by=("source_id", "variant_label", "grid_label"),
                constraints=(
                    RequireOverlappingTimerange(group_by=("instance_id",)),
                    AddParentDataset.from_defaults(SourceDatasetType.CMIP7),
                    RequireContiguousTimerange(group_by=("instance_id",)),
                    RequireFacets(
                        "variable_id",
                        required_facets=variables,
                        group_by=("source_id", "variant_label", "grid_label", "experiment_id"),
                    ),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP7),
                ),
            ),
        ),
    )
    facets = ("grid_label", "member_id", "source_id", "region", "metric")
    series = (
        SeriesDefinition(
            file_pattern="ecs/calculate/ecs_regression_*.nc",
            dimensions={
                "statistic": ("global annual mean anomaly of rtnt vs tas"),
            },
            values_name="rtnt_anomaly",
            index_name="tas_anomaly",
            attributes=[],
        ),
    )
    files = (
        FileDefinition(
            file_pattern="plots/ecs/calculate/*.png",
            dimensions={"statistic": "global annual mean anomaly of rtnt vs tas"},
        ),
        FileDefinition(
            file_pattern="work/ecs/calculate/ecs.nc",
            dimensions={"metric": "ecs"},
        ),
        FileDefinition(
            file_pattern="work/ecs/calculate/lambda.nc",
            dimensions={"metric": "lambda"},
        ),
    )

    test_data_spec = TestDataSpecification(
        test_cases=(
            TestCase(
                name="cmip6",
                description="Test with CMIP6 data.",
                requests=(
                    CMIP6Request(
                        slug="cmip6",
                        facets={
                            "experiment_id": ["abrupt-4xCO2", "piControl"],
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", "rlut", "rsdt", "rsut", "tas"],
                            "frequency": ["fx", "mon"],
                        },
                        remove_ensembles=True,
                    ),
                ),
            ),
            TestCase(
                name="cmip7",
                description="Test with CMIP7 data.",
                requests=(
                    CMIP7Request(
                        slug="cmip7",
                        facets={
                            "experiment_id": ["abrupt-4xCO2", "piControl"],
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", "rlut", "rsdt", "rsut", "tas"],
                            "branded_variable": [
                                "areacella_ti-u-hxy-u",
                                "rlut_tavg-u-hxy-u",
                                "rsdt_tavg-u-hxy-u",
                                "rsut_tavg-u-hxy-u",
                                "tas_tavg-h2m-hxy-u",
                            ],
                            "variant_label": "r1i1p1f1",
                            "frequency": ["fx", "mon"],
                            "region": "glb",
                        },
                        remove_ensembles=True,
                    ),
                ),
            ),
        )
    )

    @staticmethod
    def update_recipe(
        recipe: Recipe,
        input_files: dict[SourceDatasetType, pandas.DataFrame],
    ) -> None:
        """Update the recipe."""
        # Only run the diagnostic that computes ECS for a single model.
        recipe["diagnostics"] = {
            "ecs": {
                "description": "Calculate ECS.",
                "variables": {
                    "tas": {
                        "preprocessor": "spatial_mean",
                    },
                    "rtnt": {
                        "preprocessor": "spatial_mean",
                        "derive": True,
                    },
                },
                "scripts": {
                    "calculate": {
                        "script": "climate_metrics/ecs.py",
                        "calculate_mmm": False,
                    },
                },
            },
        }

        # Prepare updated datasets section in recipe. It contains two
        # datasets, one for the "abrupt-4xCO2" and one for the "piControl"
        # experiment.
        cmip_source = get_cmip_source_type(input_files)
        if cmip_source == SourceDatasetType.CMIP6:
            df = input_files[SourceDatasetType.CMIP6]
            recipe["datasets"] = get_child_and_parent_dataset(
                df[df.variable_id == "tas"],
                parent_experiment="piControl",
                child_duration_in_years=150,
                parent_offset_in_years=0,
                parent_duration_in_years=150,
            )
        else:
            # CMIP7: use per-variable additional_datasets to preserve correct branding_suffix
            recipe_variables = dataframe_to_recipe(
                input_files[cmip_source],
                equalize_timerange=True,
            )
            recipe["datasets"] = []
            for var_name, var_settings in recipe["diagnostics"]["ecs"]["variables"].items():
                short_name = var_settings.get("short_name", var_name)
                if short_name in recipe_variables:
                    var_settings["additional_datasets"] = recipe_variables[short_name]["additional_datasets"]
                elif var_name == "rtnt":
                    # rtnt is derived from rlut, rsdt, rsut - use rlut's dataset
                    var_settings["additional_datasets"] = recipe_variables["rlut"]["additional_datasets"]

        # Remove keys from the recipe that are only used for YAML anchors
        keys_to_remove = [
            "CMIP5_RTMT",
            "CMIP6_RTMT",
            "CMIP5_RTNT",
            "CMIP6_RTNT",
            "ECS_SCRIPT",
            "SCATTERPLOT",
        ]
        for key in keys_to_remove:
            recipe.pop(key, None)

    @staticmethod
    def format_result(
        result_dir: Path,
        execution_dataset: ExecutionDatasetCollection,
        metric_args: MetricBundleArgs,
        output_args: OutputBundleArgs,
    ) -> tuple[CMECMetric, CMECOutput]:
        """Format the result."""
        ecs_ds = xarray.open_dataset(result_dir / "work" / "ecs" / "calculate" / "ecs.nc")
        ecs = float(fillvalues_to_nan(ecs_ds["ecs"].values)[0])
        lambda_ds = xarray.open_dataset(result_dir / "work" / "ecs" / "calculate" / "lambda.nc")
        lambda_ = float(fillvalues_to_nan(lambda_ds["lambda"].values)[0])

        # Update the diagnostic bundle arguments with the computed diagnostics.
        metric_args[MetricCV.DIMENSIONS.value] = {
            MetricCV.JSON_STRUCTURE.value: [
                "region",
                "metric",
            ],
            "region": {"global": {}},
            "metric": {"ecs": {}, "lambda": {}},
        }
        metric_args[MetricCV.RESULTS.value] = {
            "global": {
                "ecs": ecs,
                "lambda": lambda_,
            },
        }

        return CMECMetric.model_validate(metric_args), CMECOutput.model_validate(output_args)



import pandas

from climate_ref_core.constraints import AddSupplementaryDataset, RequireContiguousTimerange
from climate_ref_core.datasets import FacetFilter, SourceDatasetType
from climate_ref_core.diagnostics import DataRequirement
from climate_ref_core.esgf import CMIP6Request, CMIP7Request
from climate_ref_core.metric_values.typing import FileDefinition, SeriesDefinition
from climate_ref_core.testing import TestCase, TestDataSpecification
from climate_ref_esmvaltool.diagnostics.base import ESMValToolDiagnostic, get_cmip_source_type
from climate_ref_esmvaltool.recipe import dataframe_to_recipe
from climate_ref_esmvaltool.types import Recipe


class GlobalMeanTimeseries(ESMValToolDiagnostic):
    """
    Calculate the annual mean global mean timeseries for a dataset.
    """

    name = "Global Mean Timeseries"
    slug = "global-mean-timeseries"
    base_recipe = "examples/recipe_python.yml"

    data_requirements = (
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP6,
                filters=(FacetFilter(facets={"variable_id": ("tas",)}),),
                group_by=("source_id", "experiment_id", "member_id", "table_id", "variable_id", "grid_label"),
                constraints=(
                    RequireContiguousTimerange(group_by=("instance_id",)),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP6),
                ),
            ),
        ),
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP7,
                filters=(
                    FacetFilter(
                        facets={
                            "branded_variable": "tas_tavg-h2m-hxy-u",
                            "region": "glb",
                        }
                    ),
                ),
                group_by=(
                    "source_id",
                    "experiment_id",
                    "variant_label",
                    "variable_id",
                    "grid_label",
                ),
                constraints=(
                    RequireContiguousTimerange(group_by=("instance_id",)),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP7),
                ),
            ),
        ),
    )

    facets = ()
    series = (
        SeriesDefinition(
            file_pattern="work/timeseries/script1/*.nc",
            dimensions={
                "statistic": "annual mean",
                "variable_id": "tas",
                "region": "global",
            },
            values_name="tas",
            index_name="time",
            attributes=[],
        ),
    )
    files = (
        FileDefinition(
            file_pattern="plots/timeseries/script1/png/*.png",
            dimensions={
                "statistic": "annual mean",
                "variable_id": "tas",
                "region": "global",
            },
        ),
    )

    test_data_spec = TestDataSpecification(
        test_cases=(
            TestCase(
                name="cmip6",
                description="Test with CMIP6 data.",
                requests=(
                    CMIP6Request(
                        slug="cmip6",
                        facets={
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", "tas"],
                            "frequency": ["fx", "mon"],
                            "experiment_id": "historical",
                        },
                        remove_ensembles=True,
                    ),
                ),
            ),
            TestCase(
                name="cmip7",
                description="Test with CMIP7 data.",
                requests=(
                    CMIP7Request(
                        slug="cmip7",
                        facets={
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", "tas"],
                            "branded_variable": [
                                "areacella_ti-u-hxy-u",
                                "tas_tavg-h2m-hxy-u",
                            ],
                            "variant_label": "r1i1p1f1",
                            "frequency": ["fx", "mon"],
                            "experiment_id": "historical",
                            "region": "glb",
                        },
                        remove_ensembles=True,
                    ),
                ),
            ),
        )
    )

    @staticmethod
    def update_recipe(
        recipe: Recipe,
        input_files: dict[SourceDatasetType, pandas.DataFrame],
    ) -> None:
        """Update the recipe."""
        # Clear unwanted elements from the recipe.
        recipe["datasets"].clear()
        recipe["diagnostics"].pop("map")
        variables = recipe["diagnostics"]["timeseries"]["variables"]
        variables.clear()

        # Prepare updated variables section in recipe.
        recipe_variables = dataframe_to_recipe(input_files[get_cmip_source_type(input_files)])
        recipe_variables = {k: v for k, v in recipe_variables.items() if k != "areacella"}
        for variable in recipe_variables.values():
            variable["preprocessor"] = "annual_mean_global"
            variable["caption"] = "Annual global mean {long_name} according to {dataset}."

        # Populate recipe with new variables/datasets.
        variables.update(recipe_variables)



from __future__ import annotations

import importlib.resources
from collections.abc import Iterator
from pathlib import Path
from typing import TYPE_CHECKING, Any

import cftime
import pandas as pd
import pooch
import xarray as xr
import yaml

from climate_ref_esmvaltool.types import Recipe

if TYPE_CHECKING:
    import pandas as pd


# Mapping from CMIP6 MIP table names to CMIP7 realm names.
# Base ESMValTool recipes use CMIP6 table names (Amon, Lmon, etc.) in the
# diagnostics/variables section. When running with CMIP7 data, these need
# to be rewritten to CMIP7 realm names (atmos, land, etc.).
CMIP6_MIP_TO_CMIP7_REALM = {
    "AERmon": "aerosol",
    "AERday": "aerosol",
    "Amon": "atmos",
    "Aday": "atmos",
    "CFmon": "atmos",
    "CFday": "atmos",
    "Emon": "land",
    "Eday": "land",
    "LImon": "landIce",
    "Lmon": "land",
    "Omon": "ocean",
    "Oday": "ocean",
    "OImon": "seaIce",
    "SImon": "seaIce",
    "SIday": "seaIce",
    "fx": "atmos",
    "Ofx": "ocean",
}

FACETS = {
    "CMIP6": {
        "activity": "activity_id",
        "dataset": "source_id",
        "ensemble": "member_id",
        "institute": "institution_id",
        "exp": "experiment_id",
        "grid": "grid_label",
        "mip": "table_id",
        "short_name": "variable_id",
    },
    "CMIP7": {
        "activity": "activity_id",
        "branding_suffix": "branding_suffix",
        "dataset": "source_id",
        "ensemble": "variant_label",
        "exp": "experiment_id",
        "frequency": "frequency",
        "grid": "grid_label",
        "institute": "institution_id",
        "mip": "realm",
        "region": "region",
        "short_name": "variable_id",
    },
    "obs4MIPs": {
        "dataset": "source_id",
        "frequency": "frequency",
        "grid": "grid_label",
        "institute": "institution_id",
        "short_name": "variable_id",
    },
}


def as_isodate(timestamp: pd.Timestamp) -> str:
    """Format a timestamp as an ISO 8601 datetime.

    For example, '2014-12-16 12:00:00' will be formatted as '20141216T120000'.

    Parameters
    ----------
    timestamp
        The timestamp to format.

    """
    return str(timestamp).replace(" ", "T").replace("-", "").replace(":", "")


def as_timerange(group: pd.DataFrame) -> str | None:
    """Format the timeranges from a dataframe as an ESMValTool timerange.

    Parameters
    ----------
    group
        The dataframe describing a single dataset.

    Returns
    -------
        A timerange.
    """
    # TODO: apply some rounding to avoid problems?
    # https://github.com/ESMValGroup/ESMValCore/issues/2048
    start_times = group.start_time.dropna()
    if start_times.empty:
        return None
    end_times = group.end_time.dropna()
    if end_times.empty:
        return None  # pragma: no cover
    return f"{as_isodate(start_times.min())}/{as_isodate(end_times.max())}"


def as_facets(
    group: pd.DataFrame,
) -> dict[str, Any]:
    """Convert a group from the datasets dataframe to ESMValTool facets.

    Parameters
    ----------
    group:
        A group of datasets representing a single instance_id.

    Returns
    -------
        A :obj:`dict` containing facet-value pairs.

    """
    facets = {}
    project = group.iloc[0].instance_id.split(".", 2)[0]
    facets["project"] = project
    for esmvaltool_name, ref_name in FACETS[project].items():
        values = group[ref_name].unique().tolist()
        facets[esmvaltool_name] = values if len(values) > 1 else values[0]
    timerange = as_timerange(group)
    if timerange is not None:
        facets["timerange"] = timerange
    return facets


def _iter_recipe_datasets(recipe: Recipe) -> Iterator[dict[str, Any]]:
    """Yield every dataset dict from all levels of a recipe.

    Datasets can appear at the top level (``recipe["datasets"]``),
    the diagnostic level (``diag["additional_datasets"]``), or the
    variable level (``variable["additional_datasets"]``).
    """
    yield from recipe.get("datasets", [])
    for diag in recipe.get("diagnostics", {}).values():
        yield from diag.get("additional_datasets", [])
        for var_settings in diag.get("variables", {}).values():
            if isinstance(var_settings, dict):
                yield from var_settings.get("additional_datasets", [])


def _rewrite_variable_mip(var_settings: dict[str, Any]) -> None:
    """Rewrite the mip for a single variable from a CMIP6 table name to a CMIP7 realm.

    Before overwriting the variable-level mip, the original CMIP6 value is
    pinned onto any non-CMIP7 ``additional_datasets`` (e.g. OBS) that don't
    already carry an explicit mip so they keep resolving correctly.
    """
    old_mip = var_settings.get("mip")
    if old_mip is None or old_mip not in CMIP6_MIP_TO_CMIP7_REALM:
        return

    for ds in var_settings.get("additional_datasets", []):
        if ds.get("project") != "CMIP7" and "mip" not in ds:
            ds["mip"] = old_mip

    var_settings["mip"] = CMIP6_MIP_TO_CMIP7_REALM[old_mip]


def rewrite_mip_for_cmip7(recipe: Recipe) -> None:
    """Rewrite CMIP6 MIP table names to CMIP7 realm names in a recipe.

    Base ESMValTool recipes have CMIP6 MIP table names (e.g. ``Amon``,
    ``Lmon``) hardcoded in the diagnostics/variables section. When the recipe
    uses CMIP7 data, these must be rewritten to CMIP7 realm names
    (e.g. ``atmos``, ``land``).

    Parameters
    ----------
    recipe
        The recipe to update in place.
    """
    if not any(ds.get("project") == "CMIP7" for ds in _iter_recipe_datasets(recipe)):
        return

    for diag in recipe.get("diagnostics", {}).values():
        for var_settings in diag.get("variables", {}).values():
            if isinstance(var_settings, dict):
                _rewrite_variable_mip(var_settings)


def dataframe_to_recipe(
    files: pd.DataFrame,
    group_by: tuple[str, ...] = ("instance_id",),
    equalize_timerange: bool = False,
) -> dict[str, Any]:
    """Convert the datasets dataframe to a recipe "variables" section.

    Parameters
    ----------
    files
        The pandas dataframe describing the input files.
    group_by
        The columns to group the input files by.
    equalize_timerange
        If True, use the timerange that is covered by all datasets.

    Returns
    -------
        A "variables" section that can be used in an ESMValTool recipe.
    """
    variables: dict[str, Any] = {}
    for _, group in files.groupby(list(group_by)):
        facets = as_facets(group)
        short_name = facets.pop("short_name")
        if short_name not in variables:
            variables[short_name] = {"additional_datasets": []}
        variables[short_name]["additional_datasets"].append(facets)

    if equalize_timerange:
        # Select a timerange covered by all datasets.
        start_times, end_times = [], []
        for variable in variables.values():
            for dataset in variable["additional_datasets"]:
                if "timerange" in dataset:
                    start, end = dataset["timerange"].split("/")
                    start_times.append(start)
                    end_times.append(end)
        timerange = f"{max(start_times)}/{min(end_times)}"
        for variable in variables.values():
            for dataset in variable["additional_datasets"]:
                if "timerange" in dataset:
                    dataset["timerange"] = timerange

    return variables


def get_child_and_parent_dataset(
    df: pd.DataFrame,
    parent_experiment: str,
    child_duration_in_years: int,
    parent_offset_in_years: int,
    parent_duration_in_years: int,
) -> list[dict[str, str | list[str]]]:
    """Retrieve the child and parent dataset in recipe format from a dataframe."""
    parent_df = df[(df.experiment_id == parent_experiment)]
    child_df = df[(df.experiment_id != parent_experiment)]

    if parent_df.empty:  # pragma: no branch
        raise ValueError(f"No dataset found for parent experiment '{parent_experiment}'")
    if child_df.empty:  # pragma: no branch
        raise ValueError(f"No dataset found for child experiment (not '{parent_experiment}')")

    # Compute the start time of the child and parent datasets using the
    # branch_time_in_parent and branch_time_in_child attributes to compute the offset.
    # This ensures that the datasets are aligned correctly in time.
    parent_path = parent_df.path.min()
    child_path = child_df.path.min()
    time_coder = xr.coders.CFDatetimeCoder(use_cftime=True)
    with (
        xr.open_dataset(parent_path, decode_times=time_coder) as parent_ds,
        xr.open_dataset(child_path, decode_times=time_coder) as child_ds,
    ):
        branch_time_in_parent = cftime.num2date(
            child_ds.attrs["branch_time_in_parent"],
            units=parent_ds.time.encoding["units"],
            calendar=parent_ds.time.encoding.get("calendar", "standard"),
        )
        branch_time_in_child = cftime.num2date(
            child_ds.attrs["branch_time_in_child"],
            units=child_ds.time.encoding["units"],
            calendar=child_ds.time.encoding.get("calendar", "standard"),
        )
        child_start = child_ds.time.values[0]
        parent_start = branch_time_in_parent - branch_time_in_child + child_start

    # Create the datasets for use in the recipe.
    var_name = child_df.variable_id.iloc[0]
    child_dataset = dataframe_to_recipe(child_df)[var_name]["additional_datasets"][0]
    # The end year of the timerange is inclusive, so subtract 1.
    child_end_year = child_start.year + child_duration_in_years - 1
    child_dataset["timerange"] = f"{child_start.year:04d}/{child_end_year:04d}"

    parent_dataset = dataframe_to_recipe(parent_df)[var_name]["additional_datasets"][0]
    parent_start_year = parent_start.year + parent_offset_in_years
    parent_end_year = parent_start_year + parent_duration_in_years - 1
    parent_dataset["timerange"] = f"{parent_start_year:04d}/{parent_end_year:04d}"

    return [child_dataset, parent_dataset]


_ESMVALTOOL_COMMIT = "f5214c9242725fe9a4c3628f304917c7434b361d"
_ESMVALTOOL_VERSION = f"2.13.0.dev65+g{_ESMVALTOOL_COMMIT[:9]}"
_ESMVALTOOL_URL = f"git+https://github.com/ESMValGroup/ESMValTool.git@{_ESMVALTOOL_COMMIT}"

_ESMVALCORE_COMMIT = "da81d5f67158f3d2603831b56ab6b4fb8a388d86"
_ESMVALCORE_URL = f"git+https://github.com/ESMValGroup/ESMValCore.git@{_ESMVALCORE_COMMIT}"

_RECIPES = pooch.create(
    path=pooch.os_cache("climate_ref_esmvaltool"),
    # TODO: use a released version
    # base_url="https://raw.githubusercontent.com/ESMValGroup/ESMValTool/refs/tags/v{version}/esmvaltool/recipes/",
    # version=_ESMVALTOOL_VERSION,
    base_url=f"https://raw.githubusercontent.com/ESMValGroup/ESMValTool/{_ESMVALTOOL_COMMIT}/esmvaltool/recipes/",
    env="REF_METRICS_ESMVALTOOL_DATA_DIR",
    retry_if_failed=10,
)
with importlib.resources.files("climate_ref_esmvaltool").joinpath("recipes.txt").open("rb") as _buffer:
    _RECIPES.load_registry(_buffer)


def fix_annual_statistics_keep_year(recipe: Recipe) -> None:
    """Add ``keep_group_coordinates: true`` to every ``annual_statistics`` step.

    ESMValCore changed ``annual_statistics`` to remove the ``year``
    coordinate by default (``keep_group_coordinates=False``).  Several
    ESMValTool diagnostic scripts still rely on the coordinate being
    present, so we patch the recipe to preserve it.

    Remove this workaround once ESMValCore restores the old default or
    the affected diagnostic scripts are updated.

    Parameters
    ----------
    recipe
        The recipe to update in place.
    """
    for preprocessor in recipe.get("preprocessors", {}).values():
        if isinstance(preprocessor, dict) and "annual_statistics" in preprocessor:
            annual = preprocessor["annual_statistics"]
            if isinstance(annual, dict):
                annual.setdefault("keep_group_coordinates", True)
            else:
                preprocessor["annual_statistics"] = {
                    "operator": annual if isinstance(annual, str) else "mean",
                    "keep_group_coordinates": True,
                }


def load_recipe(recipe: str) -> Recipe:
    """Load a recipe.

    Parameters
    ----------
    recipe
        The name of an ESMValTool recipe.

    Returns
    -------
        The loaded recipe.
    """
    filename = _RECIPES.fetch(recipe)

    def normalize(obj: Any) -> Any:
        # Ensure objects in the recipe are not shared.
        if isinstance(obj, dict):
            return {k: normalize(v) for k, v in obj.items()}
        if isinstance(obj, list):
            return [normalize(item) for item in obj]
        return obj

    return normalize(yaml.safe_load(Path(filename).read_text(encoding="utf-8")))  # type: ignore[no-any-return]


def prepare_climate_data(datasets: pd.DataFrame, climate_data_dir: Path) -> None:
    """Symlink the input files from the Pandas dataframe into a directory tree.

    This ensures that ESMValTool can find the data and only uses the
    requested data.

    Parameters
    ----------
    datasets
        The pandas dataframe describing the input datasets.
    climate_data_dir
        The directory where ESMValTool should look for input data.
    """
    # Track which directories we've already cleaned to avoid redundant work
    cleaned_dirs: set[Path] = set()

    for row in datasets.itertuples():
        if not isinstance(row.instance_id, str):  # pragma: no branch
            msg = f"Invalid instance_id encountered in {row}"
            raise ValueError(msg)
        if not isinstance(row.path, str):  # pragma: no branch
            msg = f"Invalid path encountered in {row}"
            raise ValueError(msg)
        if row.instance_id.startswith("obs4MIPs."):
            version = row.instance_id.split(".")[-1]
            subdirs: list[str] = ["obs4MIPs", row.source_id, version]  # type: ignore[list-item]
        elif row.instance_id.startswith("CMIP7."):
            subdirs = row.instance_id.split(".")
        else:
            subdirs = row.instance_id.split(".")
        tgt = climate_data_dir.joinpath(*subdirs) / Path(row.path).name
        tgt.parent.mkdir(parents=True, exist_ok=True)

        # Remove any stale symlinks in the target directory to prevent
        # ESMValCore from finding dangling symlinks from previous runs
        if tgt.parent not in cleaned_dirs:
            for existing in tgt.parent.iterdir():
                if existing.is_symlink() and not existing.resolve().exists():
                    existing.unlink()
            cleaned_dirs.add(tgt.parent)

        if tgt.is_symlink() or tgt.exists():
            tgt.unlink()
        tgt.symlink_to(row.path)



from collections.abc import Collection
from functools import partial

import pandas

from climate_ref_core.constraints import (
    AddSupplementaryDataset,
    PartialDateTime,
    RequireFacets,
    RequireTimerange,
)
from climate_ref_core.datasets import FacetFilter, SourceDatasetType
from climate_ref_core.diagnostics import DataRequirement
from climate_ref_core.esgf import CMIP6Request, CMIP7Request
from climate_ref_core.metric_values.typing import FileDefinition
from climate_ref_core.testing import TestCase, TestDataSpecification
from climate_ref_esmvaltool.diagnostics.base import ESMValToolDiagnostic, get_cmip_source_type
from climate_ref_esmvaltool.recipe import dataframe_to_recipe
from climate_ref_esmvaltool.types import Recipe


def get_cmip_data_requirements(
    variables: tuple[str, ...],
    branded_variables: tuple[str, ...] | None = None,
) -> tuple[tuple[DataRequirement, ...], ...]:
    """Create data requirements for CMIP6 and CMIP7 data."""
    cmip7_facets: dict[str, str | Collection[str]] = {
        "experiment_id": "historical",
        "frequency": "mon",
        "region": "glb",
        "realm": "atmos",
    }
    if branded_variables is not None:
        cmip7_facets["branded_variable"] = branded_variables
    return (
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP6,
                filters=(
                    FacetFilter(
                        facets={
                            "variable_id": variables,
                            "experiment_id": "historical",
                            "table_id": "Amon",
                        },
                    ),
                ),
                group_by=("source_id", "experiment_id", "member_id", "frequency", "grid_label"),
                constraints=(
                    RequireTimerange(
                        group_by=("instance_id",),
                        start=PartialDateTime(1996, 1),
                        end=PartialDateTime(2014, 12),
                    ),
                    RequireFacets("variable_id", variables),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP6),
                ),
            ),
        ),
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP7,
                filters=(
                    FacetFilter(
                        facets=cmip7_facets,
                    ),
                ),
                group_by=("source_id", "experiment_id", "variant_label", "frequency", "grid_label"),
                constraints=(
                    RequireTimerange(
                        group_by=("instance_id",),
                        start=PartialDateTime(1996, 1),
                        end=PartialDateTime(2014, 12),
                    ),
                    RequireFacets("variable_id", variables),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP7),
                ),
            ),
        ),
    )


def update_recipe(
    recipe: Recipe,
    input_files: dict[SourceDatasetType, pandas.DataFrame],
    var_x: str,
    var_y: str,
) -> None:
    """Update the recipe."""
    cmip_source = get_cmip_source_type(input_files)
    recipe_variables = dataframe_to_recipe(input_files[cmip_source])
    diagnostics = recipe["diagnostics"]
    diagnostic_name = f"plot_joint_{var_x}_{var_y}_model"
    diagnostic = diagnostics.pop(diagnostic_name)
    diagnostics.clear()
    diagnostics[diagnostic_name] = diagnostic
    recipe_variables = {k: v for k, v in recipe_variables.items() if k != "areacella"}
    datasets = next(iter(recipe_variables.values()))["additional_datasets"]
    for var_data in recipe_variables.values():
        for ds in var_data["additional_datasets"]:
            ds["timerange"] = "1996/2014"

    if cmip_source == SourceDatasetType.CMIP7:
        # CMIP7: use per-variable additional_datasets to preserve correct branding_suffix
        for var_name, var_settings in diagnostic.get("variables", {}).items():
            short_name = var_settings.get("short_name", var_name)
            if short_name in recipe_variables:
                var_settings["additional_datasets"] = recipe_variables[short_name]["additional_datasets"]
            else:
                # For derived variables, use the first available dataset
                var_settings["additional_datasets"] = datasets
        # Remove diagnostic-level additional_datasets to avoid CMIP6 reference datasets
        # being merged with CMIP7 per-variable datasets
        diagnostic.pop("additional_datasets", None)
    else:
        diagnostic["additional_datasets"] = datasets

    project = datasets[0]["project"]
    suptitle = f"{project} {{dataset}} {{ensemble}} {{grid}} {{timerange}}".format(**datasets[0])
    diagnostic["scripts"]["plot"]["suptitle"] = suptitle
    diagnostic["scripts"]["plot"]["plot_filename"] = (
        f"jointplot_{var_x}_{var_y}_{suptitle.replace(' ', '_').replace('/', '-')}"
    )


class CloudScatterplotCltSwcre(ESMValToolDiagnostic):
    """
    Scatterplot of clt vs swcre.
    """

    name = "Scatterplots of two cloud-relevant variables (clt vs swcre)"
    slug = "cloud-scatterplots-clt-swcre"
    base_recipe = "ref/recipe_ref_scatterplot.yml"
    facets = ()
    data_requirements = get_cmip_data_requirements(
        ("clt", "rsut", "rsutcs"),
        branded_variables=("clt_tavg-u-hxy-u", "rsut_tavg-u-hxy-u", "rsutcs_tavg-u-hxy-u"),
    )
    update_recipe = partial(update_recipe, var_x="clt", var_y="swcre")
    files = (
        FileDefinition(
            file_pattern="plots/plot_joint_clt_swcre_model/plot/png/*.png",
            dimensions={"statistic": "joint histogram of clt vs swcre"},
        ),
        FileDefinition(
            file_pattern="work/plot_joint_clt_swcre_model/plot/*.nc",
            dimensions={"statistic": "joint histogram of clt vs swcre"},
        ),
    )
    test_data_spec = TestDataSpecification(
        test_cases=(
            TestCase(
                name="cmip6",
                description="Test with CMIP6 data.",
                requests=(
                    CMIP6Request(
                        slug="cmip6",
                        facets={
                            "experiment_id": "historical",
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", "clt", "rsut", "rsutcs"],
                            "frequency": ["fx", "mon"],
                        },
                        remove_ensembles=True,
                        time_span=("1996", "2014"),
                    ),
                ),
            ),
            TestCase(
                name="cmip7",
                description="Test with CMIP7 data.",
                requests=(
                    CMIP7Request(
                        slug="cmip7",
                        facets={
                            "experiment_id": "historical",
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", "clt", "rsut", "rsutcs"],
                            "branded_variable": [
                                "areacella_ti-u-hxy-u",
                                "clt_tavg-u-hxy-u",
                                "rsut_tavg-u-hxy-u",
                                "rsutcs_tavg-u-hxy-u",
                            ],
                            "variant_label": "r1i1p1f1",
                            "frequency": ["fx", "mon"],
                            "region": "glb",
                        },
                        remove_ensembles=True,
                        time_span=("1996", "2014"),
                    ),
                ),
            ),
        )
    )


class CloudScatterplotClwviPr(ESMValToolDiagnostic):
    """
    Scatterplot of clwvi vs pr.
    """

    name = "Scatterplots of two cloud-relevant variables (clwvi vs pr)"
    slug = "cloud-scatterplots-clwvi-pr"
    base_recipe = "ref/recipe_ref_scatterplot.yml"
    facets = ()
    data_requirements = get_cmip_data_requirements(
        ("clwvi", "pr"),
        branded_variables=("clwvi_tavg-u-hxy-u", "pr_tavg-u-hxy-u"),
    )
    update_recipe = partial(update_recipe, var_x="clwvi", var_y="pr")
    files = (
        FileDefinition(
            file_pattern="plots/plot_joint_clwvi_pr_model/plot/png/*.png",
            dimensions={"statistic": "joint histogram of clwvi vs pr"},
        ),
        FileDefinition(
            file_pattern="work/plot_joint_clwvi_pr_model/plot/*.nc",
            dimensions={"statistic": "joint histogram of clwvi vs pr"},
        ),
    )
    test_data_spec = TestDataSpecification(
        test_cases=(
            TestCase(
                name="cmip6",
                description="Test with CMIP6 data.",
                requests=(
                    CMIP6Request(
                        slug="cmip6",
                        facets={
                            "experiment_id": "historical",
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", "clwvi", "pr"],
                            "frequency": ["fx", "mon"],
                        },
                        remove_ensembles=True,
                        time_span=("1996", "2014"),
                    ),
                ),
            ),
            TestCase(
                name="cmip7",
                description="Test with CMIP7 data.",
                requests=(
                    CMIP7Request(
                        slug="cmip7",
                        facets={
                            "experiment_id": "historical",
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", "clwvi", "pr"],
                            "branded_variable": [
                                "areacella_ti-u-hxy-u",
                                "clwvi_tavg-u-hxy-u",
                                "pr_tavg-u-hxy-u",
                            ],
                            "variant_label": "r1i1p1f1",
                            "frequency": ["fx", "mon"],
                            "region": "glb",
                        },
                        remove_ensembles=True,
                        time_span=("1996", "2014"),
                    ),
                ),
            ),
        )
    )


class CloudScatterplotCliviLwcre(ESMValToolDiagnostic):
    """
    Scatterplot of clivi vs lwcre.
    """

    name = "Scatterplots of two cloud-relevant variables (clivi vs lwcre)"
    slug = "cloud-scatterplots-clivi-lwcre"
    base_recipe = "ref/recipe_ref_scatterplot.yml"
    facets = ()
    data_requirements = get_cmip_data_requirements(
        ("clivi", "rlut", "rlutcs"),
        branded_variables=("clivi_tavg-u-hxy-u", "rlut_tavg-u-hxy-u", "rlutcs_tavg-u-hxy-u"),
    )
    update_recipe = partial(update_recipe, var_x="clivi", var_y="lwcre")
    files = (
        FileDefinition(
            file_pattern="plots/plot_joint_clivi_lwcre_model/plot/png/*.png",
            dimensions={"statistic": "joint histogram of clivi vs lwcre"},
        ),
        FileDefinition(
            file_pattern="work/plot_joint_clivi_lwcre_model/plot/*.nc",
            dimensions={"statistic": "joint histogram of clivi vs lwcre"},
        ),
    )
    test_data_spec = TestDataSpecification(
        test_cases=(
            TestCase(
                name="cmip6",
                description="Test with CMIP6 data.",
                requests=(
                    CMIP6Request(
                        slug="cmip6",
                        facets={
                            "experiment_id": "historical",
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", "clivi", "rlut", "rlutcs"],
                            "frequency": ["fx", "mon"],
                        },
                        remove_ensembles=True,
                        time_span=("1996", "2014"),
                    ),
                ),
            ),
            TestCase(
                name="cmip7",
                description="Test with CMIP7 data.",
                requests=(
                    CMIP7Request(
                        slug="cmip7",
                        facets={
                            "experiment_id": "historical",
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", "clivi", "rlut", "rlutcs"],
                            "branded_variable": [
                                "areacella_ti-u-hxy-u",
                                "clivi_tavg-u-hxy-u",
                                "rlut_tavg-u-hxy-u",
                                "rlutcs_tavg-u-hxy-u",
                            ],
                            "variant_label": "r1i1p1f1",
                            "frequency": ["fx", "mon"],
                            "region": "glb",
                        },
                        remove_ensembles=True,
                        time_span=("1996", "2014"),
                    ),
                ),
            ),
        )
    )


class CloudScatterplotCliTa(ESMValToolDiagnostic):
    """
    Scatterplot of cli vs ta.
    """

    name = "Scatterplots of two cloud-relevant variables (cli vs ta)"
    slug = "cloud-scatterplots-cli-ta"
    base_recipe = "ref/recipe_ref_scatterplot.yml"
    facets = ()
    data_requirements = get_cmip_data_requirements(
        ("cli", "ta"),
        branded_variables=("cli_tavg-al-hxy-u", "ta_tavg-p19-hxy-air"),
    )
    update_recipe = partial(update_recipe, var_x="cli", var_y="ta")
    files = (
        FileDefinition(
            file_pattern="plots/plot_joint_cli_ta_model/plot/png/*.png",
            dimensions={"statistic": "joint histogram of cli vs ta"},
        ),
        FileDefinition(
            file_pattern="work/plot_joint_cli_ta_model/plot/*.nc",
            dimensions={"statistic": "joint histogram of cli vs ta"},
        ),
    )
    test_data_spec = TestDataSpecification(
        test_cases=(
            TestCase(
                name="cmip6",
                description="Test with CMIP6 data.",
                requests=(
                    CMIP6Request(
                        slug="cmip6",
                        facets={
                            "experiment_id": "historical",
                            "source_id": "CESM2",
                            "variable_id": ["areacella", "cli", "ta"],
                            "frequency": ["fx", "mon"],
                        },
                        remove_ensembles=True,
                        time_span=("1996", "2014"),
                    ),
                ),
            ),
            TestCase(
                name="cmip7",
                description="Test with CMIP7 data.",
                requests=(
                    CMIP7Request(
                        slug="cmip7",
                        facets={
                            "experiment_id": "historical",
                            "source_id": "CESM2",
                            "variable_id": ["areacella", "cli", "ta"],
                            "table_id": ["fx", "Amon"],
                            "branded_variable": [
                                "areacella_ti-u-hxy-u",
                                "cli_tavg-al-hxy-u",
                                "ta_tavg-p19-hxy-air",
                            ],
                            "variant_label": "r1i1p1f1",
                            "frequency": ["fx", "mon"],
                            "region": "glb",
                        },
                        remove_ensembles=True,
                        time_span=("1996", "2014"),
                    ),
                ),
            ),
        )
    )


class CloudScatterplotsReference(ESMValToolDiagnostic):
    """
    Reference scatterplots of two cloud-relevant variables.
    """

    name = "Reference scatterplots of two cloud-relevant variables"
    slug = "cloud-scatterplots-reference"
    base_recipe = "ref/recipe_ref_scatterplot.yml"
    facets = ()
    files = (
        FileDefinition(
            file_pattern="plots/plot_joint_cli_ta_ref/plot/png/*.png",
            dimensions={"statistic": "joint histogram of cli vs ta"},
        ),
        FileDefinition(
            file_pattern="plots/plot_joint_clivi_lwcre_ref/plot/png/*.png",
            dimensions={"statistic": "joint histogram of clivi vs lwcre"},
        ),
        FileDefinition(
            file_pattern="plots/plot_joint_clt_swcre_ref/plot/png/*.png",
            dimensions={"statistic": "joint histogram of clt vs swcre"},
        ),
        FileDefinition(
            file_pattern="plots/plot_joint_clwvi_pr_ref/plot/png/*.png",
            dimensions={"statistic": "joint histogram of clwvi vs pr"},
        ),
        FileDefinition(
            file_pattern="work/plot_joint_cli_ta_ref/plot/*.nc",
            dimensions={"statistic": "joint histogram of cli vs ta"},
        ),
        FileDefinition(
            file_pattern="work/plot_joint_clivi_lwcre_ref/plot/*.nc",
            dimensions={"statistic": "joint histogram of clivi vs lwcre"},
        ),
        FileDefinition(
            file_pattern="work/plot_joint_clt_swcre_ref/plot/*.nc",
            dimensions={"statistic": "joint histogram of clt vs swcre"},
        ),
        FileDefinition(
            file_pattern="work/plot_joint_clwvi_pr_ref/plot/*.nc",
            dimensions={"statistic": "joint histogram of clwvi vs pr"},
        ),
    )
    data_requirements = (
        DataRequirement(
            source_type=SourceDatasetType.obs4MIPs,
            filters=(
                FacetFilter(
                    facets={
                        "source_id": ("ERA-5",),
                        "variable_id": ("ta",),
                    },
                ),
            ),
            group_by=("instance_id",),
            constraints=(
                RequireTimerange(
                    group_by=("instance_id",),
                    start=PartialDateTime(2007, 1),
                    end=PartialDateTime(2014, 12),
                ),
            ),
            # TODO: Add obs4MIPs datasets once available and working:
            #
            # obs4MIPs datasets with issues:
            # - GPCP-V2.3: pr
            # - CERES-EBAF-4-2: rlut, rlutcs, rsut, rsutcs
            #
            # Unsure if available on obs4MIPs:
            # - AVHRR-AMPM-fv3.0: clivi, clwvi
            # - ESACCI-CLOUD: clt
            # - CALIPSO-ICECLOUD: cli
            #
            # Related issues:
            # - https://github.com/Climate-REF/climate-ref/issues/260
            # - https://github.com/esMValGroup/esMValCore/issues/2712
            # - https://github.com/esMValGroup/esMValCore/issues/2711
            # - https://github.com/sciTools/iris/issues/6411
        ),
    )

    @staticmethod
    def update_recipe(
        recipe: Recipe,
        input_files: dict[SourceDatasetType, pandas.DataFrame],
    ) -> None:
        """Update the recipe."""
        recipe_variables = dataframe_to_recipe(input_files[SourceDatasetType.obs4MIPs])
        recipe["diagnostics"] = {k: v for k, v in recipe["diagnostics"].items() if k.endswith("_ref")}

        era5_dataset = recipe_variables["ta"]["additional_datasets"][0]
        era5_dataset["timerange"] = "2007/2015"  # Use the same timerange as for the other variable.
        era5_dataset["alias"] = era5_dataset["dataset"]
        diagnostic = recipe["diagnostics"]["plot_joint_cli_ta_ref"]
        diagnostic["variables"]["ta"]["additional_datasets"] = [era5_dataset]
        suptitle = "CALIPSO-ICECLOUD / {dataset} {timerange}".format(**era5_dataset)
        diagnostic["scripts"]["plot"]["suptitle"] = suptitle
        diagnostic["scripts"]["plot"]["plot_filename"] = (
            f"jointplot_cli_ta_{suptitle.replace(' ', '_').replace('/', '-')}"
        )

        # Use the correct obs4MIPs dataset name for dataset that cannot be ingested
        # https://github.com/Climate-REF/climate-ref/issues/260.
        diagnostic = recipe["diagnostics"]["plot_joint_clwvi_pr_ref"]
        diagnostic["variables"]["pr"]["additional_datasets"] = [
            {
                "dataset": "GPCP-V2.3",
                "project": "obs4MIPs",
                "alias": "GPCP-SG",
                "timerange": "1992/2016",
            }
        ]



import climate_ref_esmvaltool
import pytest

from climate_ref_core.datasets import DatasetCollection, ExecutionDatasetCollection, SourceDatasetType


@pytest.fixture
def execution_dataset(cmip6_data_catalog) -> ExecutionDatasetCollection:
    selected_dataset = cmip6_data_catalog[
        cmip6_data_catalog["instance_id"] == cmip6_data_catalog.instance_id.iloc[0]
    ]
    return ExecutionDatasetCollection(
        {
            SourceDatasetType.CMIP6: DatasetCollection(
                selected_dataset,
                "instance_id",
            )
        }
    )


def test_example_diagnostic(mocker, tmp_path, execution_dataset, cmip6_data_catalog, definition_factory):
    diagnostic = climate_ref_esmvaltool.provider.get("global-mean-timeseries")
    ds = cmip6_data_catalog.groupby("instance_id", as_index=False).first()

    definition = definition_factory(diagnostic=diagnostic, cmip6=DatasetCollection(ds, "instance_id"))
    definition.output_directory.mkdir(parents=True)

    result_dir = definition.output_directory / "executions" / "recipe_test_a"
    result = result_dir / "work" / "timeseries" / "script1" / "result.nc"

    def mock_run_fn(cmd, *args, **kwargs):
        result.parent.mkdir(parents=True)
        result.touch()

    mock_run = mocker.patch.object(
        climate_ref_esmvaltool.provider,
        "run",
        autospec=True,
        spec_set=True,
        side_effect=mock_run_fn,
    )

    result = diagnostic.run(definition)

    mock_run.assert_called_with(
        [
            "esmvaltool",
            "run",
            f"--config-dir={definition.to_output_path('config')}",
            f"{definition.to_output_path('recipe.yml')}",
        ],
    )

    output_bundle_path = definition.output_directory / result.output_bundle_filename
    metric_bundle_path = definition.output_directory / result.metric_bundle_filename

    assert result.successful
    assert output_bundle_path.exists()
    assert output_bundle_path.is_file()
    assert result.output_bundle_filename.name == "output.json"

    assert metric_bundle_path.exists()
    assert metric_bundle_path.is_file()
    assert result.metric_bundle_filename.name == "diagnostic.json"



"""
Regression test for ESMValTool solver output.

Uses pre-generated parquet catalogs to verify solver produces consistent results.
"""

from __future__ import annotations

import pytest
from climate_ref_esmvaltool import provider as esmvaltool_provider

from climate_ref.config import Config
from climate_ref.solve_helpers import solve_results_for_regression, solve_to_results


@pytest.fixture(scope="module")
def esmvaltool_results(esgf_data_catalog):
    esmvaltool_provider.configure(Config.default())
    return solve_to_results(esgf_data_catalog, providers=[esmvaltool_provider])


@pytest.mark.parametrize(
    "diagnostic_slug",
    [d.slug for d in esmvaltool_provider.diagnostics()],
)
def test_solve_regression(esmvaltool_results, data_regression, diagnostic_slug):
    filtered = [r for r in esmvaltool_results if r["diagnostic"] == diagnostic_slug]
    regression = solve_results_for_regression(filtered)
    data_regression.check(regression)



"""ESMValTool diagnostics."""

from climate_ref_esmvaltool.diagnostics.climate_at_global_warming_levels import ClimateAtGlobalWarmingLevels
from climate_ref_esmvaltool.diagnostics.climate_drivers_for_fire import ClimateDriversForFire
from climate_ref_esmvaltool.diagnostics.cloud_radiative_effects import CloudRadiativeEffects
from climate_ref_esmvaltool.diagnostics.cloud_scatterplots import (
    CloudScatterplotCliTa,
    CloudScatterplotCliviLwcre,
    CloudScatterplotCltSwcre,
    CloudScatterplotClwviPr,
    CloudScatterplotsReference,
)
from climate_ref_esmvaltool.diagnostics.ecs import EquilibriumClimateSensitivity
from climate_ref_esmvaltool.diagnostics.enso import ENSOBasicClimatology, ENSOCharacteristics
from climate_ref_esmvaltool.diagnostics.example import GlobalMeanTimeseries
from climate_ref_esmvaltool.diagnostics.regional_historical_changes import (
    RegionalHistoricalAnnualCycle,
    RegionalHistoricalTimeSeries,
    RegionalHistoricalTrend,
)
from climate_ref_esmvaltool.diagnostics.sea_ice_area_basic import SeaIceAreaBasic
from climate_ref_esmvaltool.diagnostics.sea_ice_sensitivity import SeaIceSensitivity
from climate_ref_esmvaltool.diagnostics.tcr import TransientClimateResponse
from climate_ref_esmvaltool.diagnostics.tcre import TransientClimateResponseEmissions
from climate_ref_esmvaltool.diagnostics.zec import ZeroEmissionCommitment

__all__ = [
    "ClimateAtGlobalWarmingLevels",
    "ClimateDriversForFire",
    "CloudRadiativeEffects",
    "CloudScatterplotCliTa",
    "CloudScatterplotCliviLwcre",
    "CloudScatterplotCltSwcre",
    "CloudScatterplotClwviPr",
    "CloudScatterplotsReference",
    "ENSOBasicClimatology",
    "ENSOCharacteristics",
    "EquilibriumClimateSensitivity",
    "GlobalMeanTimeseries",
    "RegionalHistoricalAnnualCycle",
    "RegionalHistoricalTimeSeries",
    "RegionalHistoricalTrend",
    "SeaIceAreaBasic",
    "SeaIceSensitivity",
    "TransientClimateResponse",
    "TransientClimateResponseEmissions",
    "ZeroEmissionCommitment",
]



from pathlib import Path

import pandas
import pandas as pd

from climate_ref_core.constraints import (
    AddSupplementaryDataset,
    RequireContiguousTimerange,
    RequireFacets,
    RequireOverlappingTimerange,
)
from climate_ref_core.datasets import ExecutionDatasetCollection, FacetFilter, SourceDatasetType
from climate_ref_core.diagnostics import DataRequirement
from climate_ref_core.esgf import CMIP6Request, CMIP7Request
from climate_ref_core.metric_values.typing import FileDefinition, SeriesDefinition
from climate_ref_core.pycmec.metric import CMECMetric, MetricCV
from climate_ref_core.pycmec.output import CMECOutput
from climate_ref_core.testing import TestCase, TestDataSpecification
from climate_ref_esmvaltool.diagnostics.base import ESMValToolDiagnostic, get_cmip_source_type
from climate_ref_esmvaltool.recipe import dataframe_to_recipe
from climate_ref_esmvaltool.types import MetricBundleArgs, OutputBundleArgs, Recipe


class ENSOBasicClimatology(ESMValToolDiagnostic):
    """
    Calculate the ENSO CLIVAR metrics - background climatology.
    """

    name = "ENSO Basic Climatology"
    slug = "enso-basic-climatology"
    base_recipe = "ref/recipe_enso_basicclimatology.yml"

    data_requirements = (
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP6,
                filters=(
                    FacetFilter(
                        facets={
                            "variable_id": ("pr", "tauu"),
                            "experiment_id": "historical",
                            "table_id": "Amon",
                        },
                    ),
                    FacetFilter(
                        facets={
                            "variable_id": "tos",
                            "experiment_id": "historical",
                            "table_id": "Omon",
                        },
                    ),
                ),
                group_by=("source_id", "member_id", "grid_label"),
                constraints=(
                    RequireContiguousTimerange(group_by=("instance_id",)),
                    RequireOverlappingTimerange(group_by=("instance_id",)),
                    RequireFacets(
                        "variable_id",
                        (
                            "pr",
                            "tauu",
                            "tos",
                        ),
                    ),
                ),
            ),
        ),
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP7,
                filters=(
                    FacetFilter(
                        facets={
                            "branded_variable": (
                                "pr_tavg-u-hxy-u",
                                "tauu_tavg-u-hxy-u",
                            ),
                            "experiment_id": "historical",
                            "frequency": "mon",
                            "region": "glb",
                            "realm": "atmos",
                        },
                    ),
                    FacetFilter(
                        facets={
                            "branded_variable": "tos_tavg-u-hxy-sea",
                            "experiment_id": "historical",
                            "frequency": "mon",
                            "region": "glb",
                            "realm": "ocean",
                        },
                    ),
                ),
                group_by=("source_id", "variant_label", "grid_label"),
                constraints=(
                    RequireContiguousTimerange(group_by=("instance_id",)),
                    RequireOverlappingTimerange(group_by=("instance_id",)),
                    RequireFacets(
                        "variable_id",
                        (
                            "pr",
                            "tauu",
                            "tos",
                        ),
                    ),
                ),
            ),
        ),
    )
    facets = ()

    files = (
        tuple(
            FileDefinition(
                file_pattern=f"plots/diagnostic_metrics/plot_script/png/*_eq_{var_name}_bias.png",
                dimensions={
                    "statistic": (
                        f"zonal bias in the time-mean {var_name} structure across the equatorial Pacific"
                    ),
                },
            )
            for var_name in ("pr", "sst", "tauu")
        )
        + tuple(
            FileDefinition(
                file_pattern=f"plots/diagnostic_metrics/plot_script/png/*_eq_{var_name}_seacycle.png",
                dimensions={
                    "statistic": (
                        "zonal bias in the amplitude of the mean seasonal cycle of "
                        f"{var_name} in the equatorial Pacific"
                    ),
                },
            )
            for var_name in ("pr", "sst", "tauu")
        )
        + (
            FileDefinition(
                file_pattern="plots/diagnostic_metrics/plot_script/png/*_pr_double.png",
                dimensions={
                    "statistic": "meridional bias in the time-mean pr structure across the eastern Pacific",
                },
            ),
            FileDefinition(
                file_pattern="plots/diagnostic_metrics/plot_script/png/*_pr_double_seacycle.png",
                dimensions={
                    "statistic": (
                        "meridional bias in the amplitude of the mean seasonal "
                        "pr cycle in the eastern Pacific"
                    ),
                },
            ),
        )
        + tuple(
            FileDefinition(
                file_pattern=f"plots/diagnostic_level2/plot_script/png/*_{var_name}_map_*.png",
                dimensions={
                    "variable_id": "tos" if var_name == "tos" else var_name,
                },
            )
            for var_name in ("pr", "tauu", "tos")
        )
    )

    series = (
        tuple(
            SeriesDefinition(
                file_pattern=f"diagnostic_metrics/plot_script/{source_id}_eq_{var_name}_bias.nc",
                dimensions=(
                    {
                        "statistic": (
                            f"zonal bias in the time-mean {var_name} structure across the equatorial Pacific"
                        ),
                    }
                    | ({} if source_id == "{source_id}" else {"reference_source_id": source_id})
                ),
                values_name="tos" if var_name == "sst" else var_name,
                index_name="lon",
                attributes=[],
            )
            for var_name in ("pr", "sst", "tauu")
            for source_id in ("{source_id}", "GPCP-V2.3", "TROPFLUX")
        )
        + tuple(
            SeriesDefinition(
                file_pattern=f"diagnostic_metrics/plot_script/{{source_id}}_eq_{var_name}_seacycle.nc",
                dimensions=(
                    {
                        "statistic": (
                            "zonal bias in the amplitude of the mean seasonal cycle of "
                            f"{var_name} in the equatorial Pacific"
                        ),
                    }
                    | ({} if source_id == "{source_id}" else {"reference_source_id": source_id})
                ),
                values_name="tos" if var_name == "sst" else var_name,
                index_name="lon",
                attributes=[],
            )
            for var_name in ("pr", "sst", "tauu")
            for source_id in ("{source_id}", "GPCP-V2.3", "TROPFLUX")
        )
        + tuple(
            SeriesDefinition(
                file_pattern="diagnostic_metrics/plot_script/{source_id}_pr_double.nc",
                dimensions=(
                    {
                        "statistic": (
                            "meridional bias in the time-mean pr structure across the eastern Pacific"
                        ),
                    }
                    | ({} if source_id == "{source_id}" else {"reference_source_id": source_id})
                ),
                values_name="pr",
                index_name="lat",
                attributes=[],
            )
            for source_id in ("{source_id}", "GPCP-V2.3")
        )
        + tuple(
            SeriesDefinition(
                file_pattern="diagnostic_metrics/plot_script/*_pr_double_seacycle.nc",
                dimensions=(
                    {
                        "statistic": (
                            "meridional bias in the amplitude of the mean seasonal "
                            "pr cycle in the eastern Pacific"
                        ),
                    }
                    | ({} if source_id == "{source_id}" else {"reference_source_id": source_id})
                ),
                values_name="pr",
                index_name="lat",
                attributes=[],
            )
            for source_id in ("{source_id}", "GPCP-V2.3")
        )
    )

    test_data_spec = TestDataSpecification(
        test_cases=(
            TestCase(
                name="cmip6",
                description="Test with CMIP6 data.",
                requests=(
                    CMIP6Request(
                        slug="cmip6",
                        facets={
                            "experiment_id": "historical",
                            "source_id": "CanESM5",
                            "variable_id": ["pr", "tauu", "tos"],
                            "frequency": "mon",
                        },
                        remove_ensembles=True,
                    ),
                ),
            ),
            TestCase(
                name="cmip7",
                description="Test with CMIP7 data.",
                requests=(
                    CMIP7Request(
                        slug="cmip7",
                        facets={
                            "experiment_id": "historical",
                            "source_id": "CanESM5",
                            "variable_id": ["pr", "tauu", "tos"],
                            "branded_variable": [
                                "pr_tavg-u-hxy-u",
                                "tauu_tavg-u-hxy-u",
                                "tos_tavg-u-hxy-sea",
                            ],
                            "variant_label": "r1i1p1f1",
                            "frequency": "mon",
                            "region": "glb",
                        },
                        remove_ensembles=True,
                    ),
                ),
            ),
        )
    )

    @staticmethod
    def update_recipe(
        recipe: Recipe,
        input_files: dict[SourceDatasetType, pandas.DataFrame],
    ) -> None:
        """Update the recipe."""
        recipe_variables = dataframe_to_recipe(input_files[get_cmip_source_type(input_files)])
        recipe.pop("datasets")
        for diagnostic in recipe["diagnostics"].values():
            for variable in diagnostic["variables"].values():
                variable["additional_datasets"].extend(
                    recipe_variables[variable["short_name"]]["additional_datasets"]
                )


class ENSOCharacteristics(ESMValToolDiagnostic):
    """
    Calculate the ENSO CLIVAR metrics - basic ENSO characteristics.
    """

    name = "ENSO Characteristics"
    slug = "enso-characteristics"
    base_recipe = "ref/recipe_enso_characteristics.yml"

    data_requirements = (
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP6,
                filters=(
                    FacetFilter(
                        facets={
                            "variable_id": "tos",
                            "experiment_id": "historical",
                            "table_id": "Omon",
                        },
                    ),
                ),
                group_by=("source_id", "member_id", "grid_label"),
                constraints=(
                    RequireContiguousTimerange(group_by=("instance_id",)),
                    RequireOverlappingTimerange(group_by=("instance_id",)),
                    AddSupplementaryDataset.from_defaults("areacello", SourceDatasetType.CMIP6),
                    RequireFacets("variable_id", ("tos", "areacello")),
                ),
            ),
        ),
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP7,
                filters=(
                    FacetFilter(
                        facets={
                            "branded_variable": "tos_tavg-u-hxy-sea",
                            "experiment_id": "historical",
                            "frequency": "mon",
                            "region": "glb",
                            "realm": "ocean",
                        },
                    ),
                ),
                group_by=("source_id", "variant_label", "grid_label"),
                constraints=(
                    RequireContiguousTimerange(group_by=("instance_id",)),
                    RequireOverlappingTimerange(group_by=("instance_id",)),
                    AddSupplementaryDataset.from_defaults("areacello", SourceDatasetType.CMIP7),
                    RequireFacets("variable_id", ("tos", "areacello")),
                ),
            ),
        ),
    )
    facets = ("grid_label", "member_id", "source_id", "region", "metric")
    # ENSO pattern and lifecycle are series, but the ESMValTool diagnostic
    # script does not save the values used in the figure.
    series = tuple()
    files = tuple(
        FileDefinition(
            file_pattern=f"plots/diagnostic_metrics/plot_script/png/*_{metric}.png",
            dimensions={"metric": metric},
        )
        for metric in (
            "09pattern",
            "10lifecycle",
            "11amplitude",
            "12seasonality",
            "13asymmetry",
            "14duration",
            "15diversity",
        )
    )

    test_data_spec = TestDataSpecification(
        test_cases=(
            TestCase(
                name="cmip6",
                description="Test with CMIP6 data.",
                requests=(
                    CMIP6Request(
                        slug="cmip6",
                        facets={
                            "experiment_id": "historical",
                            "source_id": "CanESM5",
                            "variable_id": ["areacello", "tos"],
                            "frequency": ["fx", "mon"],
                        },
                        remove_ensembles=True,
                    ),
                ),
            ),
            TestCase(
                name="cmip7",
                description="Test with CMIP7 data.",
                requests=(
                    CMIP7Request(
                        slug="cmip7",
                        facets={
                            "experiment_id": "historical",
                            "source_id": "CanESM5",
                            "variable_id": ["areacello", "tos"],
                            "branded_variable": [
                                "areacello_ti-u-hxy-u",
                                "tos_tavg-u-hxy-sea",
                            ],
                            "variant_label": "r1i1p1f1",
                            "frequency": ["fx", "mon"],
                            "region": "glb",
                        },
                        remove_ensembles=True,
                    ),
                ),
            ),
        )
    )

    @staticmethod
    def update_recipe(
        recipe: Recipe,
        input_files: dict[SourceDatasetType, pandas.DataFrame],
    ) -> None:
        """Update the recipe."""
        recipe_variables = dataframe_to_recipe(input_files[get_cmip_source_type(input_files)])
        recipe["datasets"] = recipe_variables["tos"]["additional_datasets"]
        # TODO: update the observational data requirement once available on ESGF.
        # Observations - use only one per run
        recipe["datasets"].append(
            # {
            #     "dataset": "NOAA-ERSSTv5",
            #     "version": "v5",
            #     "project": "OBS6",
            #     "type": "reanaly",
            #     "tier": 2,
            # }
            {
                "dataset": "TROPFLUX",
                "version": "v1",
                "project": "OBS6",
                "type": "reanaly",
                "tier": 2,
                "mip": "Omon",
            }
        )

    @staticmethod
    def format_result(
        result_dir: Path,
        execution_dataset: ExecutionDatasetCollection,
        metric_args: MetricBundleArgs,
        output_args: OutputBundleArgs,
    ) -> tuple[CMECMetric, CMECOutput]:
        """Format the result."""
        metrics = pd.read_csv(
            result_dir / "work" / "diagnostic_metrics" / "plot_script" / "matrix.csv",
            names=["dataset", "metric_name", "metric_value"],
        )

        # Update the diagnostic bundle arguments with the computed diagnostics.
        metric_args[MetricCV.DIMENSIONS.value] = {
            "json_structure": [
                "region",
                "metric",
            ],
            "region": {"global": {}},
            "metric": {metric: {} for metric in metrics.metric_name},
        }
        metric_args[MetricCV.RESULTS.value] = {
            "global": {
                metric_name: metric_value
                for metric_name, metric_value in zip(
                    metrics.metric_name,
                    metrics.metric_value,
                )
            },
        }

        return CMECMetric.model_validate(metric_args), CMECOutput.model_validate(output_args)



from pathlib import Path

import pandas
import pandas as pd

from climate_ref_core.constraints import (
    AddSupplementaryDataset,
    PartialDateTime,
    RequireFacets,
    RequireTimerange,
)
from climate_ref_core.datasets import ExecutionDatasetCollection, FacetFilter, SourceDatasetType
from climate_ref_core.diagnostics import DataRequirement
from climate_ref_core.esgf import CMIP6Request, CMIP7Request
from climate_ref_core.metric_values.typing import FileDefinition
from climate_ref_core.pycmec.metric import CMECMetric, MetricCV
from climate_ref_core.pycmec.output import CMECOutput
from climate_ref_core.testing import TestCase, TestDataSpecification
from climate_ref_esmvaltool.diagnostics.base import ESMValToolDiagnostic, get_cmip_source_type
from climate_ref_esmvaltool.recipe import dataframe_to_recipe
from climate_ref_esmvaltool.types import MetricBundleArgs, OutputBundleArgs, Recipe


class SeaIceSensitivity(ESMValToolDiagnostic):
    """
    Calculate sea ice sensitivity.
    """

    name = "Sea ice sensitivity"
    slug = "sea-ice-sensitivity"
    base_recipe = "recipe_seaice_sensitivity.yml"

    data_requirements = (
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP6,
                filters=(
                    FacetFilter(
                        facets={
                            "variable_id": "siconc",
                            "experiment_id": "historical",
                            "table_id": "SImon",
                        },
                    ),
                    FacetFilter(
                        facets={
                            "variable_id": "tas",
                            "experiment_id": "historical",
                            "table_id": "Amon",
                        },
                    ),
                ),
                group_by=("experiment_id",),  # this does nothing, but group_by cannot be empty
                constraints=(
                    RequireTimerange(
                        group_by=("instance_id",),
                        start=PartialDateTime(1979, 1),
                        end=PartialDateTime(2014, 12),
                    ),
                    RequireFacets(
                        "variable_id",
                        required_facets=("siconc", "tas"),
                        group_by=("source_id", "member_id", "grid_label"),
                    ),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP6),
                    AddSupplementaryDataset.from_defaults("areacello", SourceDatasetType.CMIP6),
                    RequireFacets(
                        "variable_id",
                        required_facets=("areacello",),
                        group_by=("source_id", "grid_label"),
                    ),
                ),
            ),
        ),
        (
            DataRequirement(
                source_type=SourceDatasetType.CMIP7,
                filters=(
                    FacetFilter(
                        facets={
                            "branded_variable": "siconc_tavg-u-hxy-u",
                            "experiment_id": "historical",
                            "frequency": "mon",
                            "region": "glb",
                            "realm": "seaIce",
                        },
                    ),
                    FacetFilter(
                        facets={
                            "branded_variable": "tas_tavg-h2m-hxy-u",
                            "experiment_id": "historical",
                            "frequency": "mon",
                            "region": "glb",
                            "realm": "atmos",
                        },
                    ),
                ),
                group_by=("experiment_id",),  # this does nothing, but group_by cannot be empty
                constraints=(
                    RequireTimerange(
                        group_by=("instance_id",),
                        start=PartialDateTime(1979, 1),
                        end=PartialDateTime(2014, 12),
                    ),
                    RequireFacets(
                        "variable_id",
                        required_facets=("siconc", "tas"),
                        group_by=("source_id", "variant_label", "grid_label"),
                    ),
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP7),
                    AddSupplementaryDataset.from_defaults("areacello", SourceDatasetType.CMIP7),
                    RequireFacets(
                        "variable_id",
                        required_facets=("areacello",),
                        group_by=("source_id", "grid_label"),
                    ),
                ),
            ),
        ),
    )
    facets = ("experiment_id", "source_id", "region", "metric")
    files = tuple(
        FileDefinition(
            file_pattern=f"plots/{region}/sea_ice_sensitivity_script/png/*.png",
            dimensions={"region": region},
        )
        for region in ("arctic", "antarctic")
    ) + tuple(
        FileDefinition(
            file_pattern=f"work/{region}/sea_ice_sensitivity_script/plotted_values.csv",
            dimensions={"region": region},
        )
        for region in ("arctic", "antarctic")
    )

    test_data_spec = TestDataSpecification(
        test_cases=(
            TestCase(
                name="cmip6",
                description="Test with CMIP6 data.",
                requests=(
                    CMIP6Request(
                        slug="cmip6",
                        facets={
                            "experiment_id": "historical",
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", "areacello", "siconc", "tas"],
                            "frequency": ["fx", "mon"],
                        },
                        remove_ensembles=True,
                        time_span=("1979", "2014"),
                    ),
                ),
            ),
            TestCase(
                name="cmip7",
                description="Test with CMIP7 data.",
                requests=(
                    CMIP7Request(
                        slug="cmip7",
                        facets={
                            "experiment_id": "historical",
                            "source_id": "CanESM5",
                            "variable_id": ["areacella", "areacello", "siconc", "tas"],
                            "branded_variable": [
                                "areacella_ti-u-hxy-u",
                                "areacello_ti-u-hxy-u",
                                "siconc_tavg-u-hxy-u",
                                "tas_tavg-h2m-hxy-u",
                            ],
                            "variant_label": "r1i1p1f1",
                            "frequency": ["fx", "mon"],
                            "region": "glb",
                        },
                        remove_ensembles=True,
                        time_span=("1979", "2014"),
                    ),
                ),
            ),
        )
    )

    @staticmethod
    def update_recipe(
        recipe: Recipe,
        input_files: dict[SourceDatasetType, pandas.DataFrame],
    ) -> None:
        """Update the recipe."""
        cmip_source = get_cmip_source_type(input_files)
        recipe_variables = dataframe_to_recipe(input_files[cmip_source])

        if cmip_source == SourceDatasetType.CMIP7:
            # CMIP7: use per-variable additional_datasets to preserve correct branding_suffix
            recipe["datasets"] = []
            for diagnostic in recipe["diagnostics"].values():
                for var_name, variable in diagnostic.get("variables", {}).items():
                    short_name = variable.get("short_name", var_name)
                    if short_name in recipe_variables:
                        datasets = recipe_variables[short_name]["additional_datasets"]
                        for ds in datasets:
                            ds.pop("mip", None)
                            ds["timerange"] = "1979/2014"
                        variable["additional_datasets"] = datasets
        else:
            datasets = recipe_variables["tas"]["additional_datasets"]
            for dataset in datasets:
                dataset.pop("mip")
                dataset["timerange"] = "1979/2014"
            recipe["datasets"] = datasets

    @staticmethod
    def format_result(
        result_dir: Path,
        execution_dataset: ExecutionDatasetCollection,
        metric_args: MetricBundleArgs,
        output_args: OutputBundleArgs,
    ) -> tuple[CMECMetric, CMECOutput]:
        """Format the result."""
        metric_args[MetricCV.DIMENSIONS.value] = {
            "json_structure": [
                "source_id",
                "region",
                "metric",
            ],
            "source_id": {},
            "region": {},
            "metric": {},
        }
        for region in "antarctic", "arctic":
            df = pd.read_csv(
                result_dir / "work" / region / "sea_ice_sensitivity_script" / "plotted_values.csv"
            )
            df = df.rename(columns={"Unnamed: 0": "source_id"}).drop(columns=["label"])
            metric_args[MetricCV.DIMENSIONS.value]["region"][region] = {}
            for metric in df.columns[1:]:
                metric_args[MetricCV.DIMENSIONS.value]["metric"][metric] = {}
            for row in df.itertuples(index=False):
                source_id = row.source_id
                metric_args[MetricCV.DIMENSIONS.value]["source_id"][source_id] = {}
                for metric, value in zip(df.columns[1:], row[1:]):
                    if source_id not in metric_args[MetricCV.RESULTS.value]:
                        metric_args[MetricCV.RESULTS.value][source_id] = {}
                    if region not in metric_args[MetricCV.RESULTS.value][source_id]:
                        metric_args[MetricCV.RESULTS.value][source_id][region] = {}
                    metric_args[MetricCV.RESULTS.value][source_id][region][metric] = value

        return CMECMetric.model_validate(metric_args), CMECOutput.model_validate(output_args)



import datetime
from pathlib import Path

import pandas as pd
import pytest
import xarray as xr
from climate_ref_esmvaltool.recipe import get_child_and_parent_dataset, prepare_climate_data


def test_get_child_and_parent_dataset(tmp_path):
    # Code to extract the dataframe below from the TCR diagnostic:
    # df[df.variable_id == "tas"].to_dict(orient="list")
    df = pd.DataFrame(
        {
            "instance_id": [
                "CMIP6.CMIP.CCCma.CanESM5.1pctCO2.r1i1p1f1.Amon.tas.gn.v20190429",
                "CMIP6.CMIP.CCCma.CanESM5.piControl.r1i1p1f1.Amon.tas.gn.v20190429",
            ],
            "activity_id": ["CMIP", "CMIP"],
            "branch_method": ["Spin-up documentation", "Spin-up documentation"],
            "branch_time_in_child": [0.0, 1223115.0],
            "branch_time_in_parent": [1223115.0, 1223115.0],
            "experiment": ["1 percent per year increase in CO2", "pre-industrial control"],
            "experiment_id": ["1pctCO2", "piControl"],
            "grid_label": ["gn", "gn"],
            "institution_id": ["CCCma", "CCCma"],
            "nominal_resolution": ["500 km", "500 km"],
            "parent_activity_id": ["CMIP", "CMIP"],
            "parent_experiment_id": ["piControl", "piControl-spinup"],
            "parent_source_id": ["CanESM5", "CanESM5"],
            "parent_time_units": ["days since 1850-01-01 0:0:0.0", "days since 1850-01-01 0:0:0.0"],
            "parent_variant_label": ["r1i1p1f1", "r1i1p1f1"],
            "product": ["model-output", "model-output"],
            "realm": ["atmos", "atmos"],
            "source_id": ["CanESM5", "CanESM5"],
            "source_type": ["AOGCM", "AOGCM"],
            "sub_experiment": ["none", "none"],
            "sub_experiment_id": ["none", "none"],
            "table_id": ["Amon", "Amon"],
            "variable_id": ["tas", "tas"],
            "variant_label": ["r1i1p1f1", "r1i1p1f1"],
            "member_id": ["r1i1p1f1", "r1i1p1f1"],
            "standard_name": ["air_temperature", "air_temperature"],
            "long_name": ["Near-Surface Air Temperature", "Near-Surface Air Temperature"],
            "units": ["K", "K"],
            "vertical_levels": [1, 1],
            "start_time": [datetime.datetime(1850, 1, 16, 12, 0), datetime.datetime(5201, 1, 16, 12, 0)],
            "end_time": [datetime.datetime(1989, 12, 16, 12, 0), datetime.datetime(5340, 12, 16, 12, 0)],
            "time_range": [
                "1850-01-16 12:00:00-1989-12-16 12:00:00",
                "5201-01-16 12:00:00-5340-12-16 12:00:00",
            ],
            "path": [
                f"{tmp_path}/tas_Amon_CanESM5_1pctCO2_r1i1p1f1_gn_185001-198912.nc",
                f"{tmp_path}/tas_Amon_CanESM5_piControl_r1i1p1f1_gn_520101-534012.nc",
            ],
            "version": ["v20190429", "v20190429"],
            "finalised": [True, True],
        }
    )

    # Code to generate the file content below:
    # ds = xr.open_dataset(
    #     "tas_Amon_CanESM5_1pctCO2_r1i1p1f1_gn_185001-198912.nc",
    #     decode_times=False,
    # )
    # ds.isel(time=slice(0, 1), lat=slice(0,1), lon=slice(0,1)).to_dict()
    xr.Dataset.from_dict(
        {
            "coords": {
                "time": {
                    "dims": ("time",),
                    "attrs": {
                        "bounds": "time_bnds",
                        "axis": "T",
                        "long_name": "time",
                        "standard_name": "time",
                        "units": "days since 1850-01-01",
                        "calendar": "365_day",
                    },
                    "data": [15.5],
                },
                "height": {
                    "dims": (),
                    "attrs": {
                        "units": "m",
                        "axis": "Z",
                        "positive": "up",
                        "long_name": "height",
                        "standard_name": "height",
                    },
                    "data": 2.0,
                },
                "lat": {
                    "dims": ("lat",),
                    "attrs": {
                        "bounds": "lat_bnds",
                        "units": "degrees_north",
                        "axis": "Y",
                        "long_name": "Latitude",
                        "standard_name": "latitude",
                    },
                    "data": [-90.0],
                },
                "lon": {
                    "dims": ("lon",),
                    "attrs": {
                        "bounds": "lon_bnds",
                        "units": "degrees_east",
                        "axis": "X",
                        "long_name": "Longitude",
                        "standard_name": "longitude",
                    },
                    "data": [0.0],
                },
            },
            "attrs": {
                "CCCma_model_hash": "3dedf95315d603326fde4f5340dc0519d80d10c0",
                "CCCma_parent_runid": "rc3-pictrl",
                "CCCma_pycmor_hash": "33c30511acc319a98240633965a04ca99c26427e",
                "CCCma_runid": "rc3.1-1ppy",
                "Conventions": "CF-1.7 CMIP-6.2",
                "YMDH_branch_time_in_child": "1850:01:01:00",
                "YMDH_branch_time_in_parent": "5201:01:01:00",
                "activity_id": "CMIP",
                "branch_method": "Spin-up documentation",
                "branch_time_in_child": 0.0,
                "branch_time_in_parent": 1223115.0,
                "contact": "ec.cccma.info-info.ccmac.ec@canada.ca",
                "creation_date": "2019-04-30T17:28:14Z",
                "data_specs_version": "01.00.29",
                "experiment": "1 percent per year increase in CO2",
                "experiment_id": "1pctCO2",
                "external_variables": "areacella",
                "forcing_index": 1,
                "frequency": "mon",
                "further_info_url": "https://furtherinfo.es-doc.org/CMIP6.CCCma.CanESM5.1pctCO2.none.r1i1p1f1",
                "grid_label": "gn",
                "initialization_index": 1,
                "institution_id": "CCCma",
                "mip_era": "CMIP6",
                "nominal_resolution": "500 km",
                "parent_activity_id": "CMIP",
                "parent_experiment_id": "piControl",
                "parent_mip_era": "CMIP6",
                "parent_source_id": "CanESM5",
                "parent_time_units": "days since 1850-01-01 0:0:0.0",
                "parent_variant_label": "r1i1p1f1",
                "physics_index": 1,
                "product": "model-output",
                "realization_index": 1,
                "realm": "atmos",
                "source_id": "CanESM5",
                "source_type": "AOGCM",
                "sub_experiment": "none",
                "sub_experiment_id": "none",
                "table_id": "Amon",
                "title": "CanESM5 output prepared for CMIP6",
                "tracking_id": "hdl:21.14100/3beb38c2-d30a-4ed6-8198-b7b5182cab28",
                "variable_id": "tas",
                "variant_label": "r1i1p1f1",
                "version": "v20190429",
                "cmor_version": "3.4.0",
            },
            "dims": {"time": 1, "lat": 1, "lon": 1, "bnds": 2},
            "data_vars": {
                "tas": {
                    "dims": ("time", "lat", "lon"),
                    "attrs": {
                        "standard_name": "air_temperature",
                        "long_name": "Near-Surface Air Temperature",
                        "units": "K",
                        "original_name": "ST",
                        "cell_methods": "area: time: mean",
                        "cell_measures": "area: areacella",
                    },
                    "data": [[[247.98342895507812]]],
                },
                "lat_bnds": {"dims": ("lat", "bnds"), "attrs": {}, "data": [[-90.0, -85.0]]},
                "lon_bnds": {"dims": ("lon", "bnds"), "attrs": {}, "data": [[-5.0, 5.0]]},
                "time_bnds": {"dims": ("time", "bnds"), "attrs": {}, "data": [[0.0, 31.0]]},
            },
        }
    ).to_netcdf(tmp_path / "tas_Amon_CanESM5_1pctCO2_r1i1p1f1_gn_185001-198912.nc")

    xr.Dataset.from_dict(
        {
            "coords": {
                "time": {
                    "dims": ("time",),
                    "attrs": {
                        "bounds": "time_bnds",
                        "axis": "T",
                        "long_name": "time",
                        "standard_name": "time",
                        "units": "days since 1850-01-01",
                        "calendar": "365_day",
                    },
                    "data": [1223130.5],
                },
                "height": {
                    "dims": (),
                    "attrs": {
                        "units": "m",
                        "axis": "Z",
                        "positive": "up",
                        "long_name": "height",
                        "standard_name": "height",
                    },
                    "data": 2.0,
                },
                "lat": {
                    "dims": ("lat",),
                    "attrs": {
                        "bounds": "lat_bnds",
                        "units": "degrees_north",
                        "axis": "Y",
                        "long_name": "Latitude",
                        "standard_name": "latitude",
                    },
                    "data": [-90.0],
                },
                "lon": {
                    "dims": ("lon",),
                    "attrs": {
                        "bounds": "lon_bnds",
                        "units": "degrees_east",
                        "axis": "X",
                        "long_name": "Longitude",
                        "standard_name": "longitude",
                    },
                    "data": [0.0],
                },
            },
            "attrs": {
                "CCCma_model_hash": "24718c8346665b218729640ffe79d263b76456c4",
                "CCCma_parent_runid": "rc3-pictrl",
                "CCCma_pycmor_hash": "33c30511acc319a98240633965a04ca99c26427e",
                "CCCma_runid": "rc3.1-pictrl",
                "Conventions": "CF-1.7 CMIP-6.2",
                "YMDH_branch_time_in_child": "5201:01:01:00",
                "YMDH_branch_time_in_parent": "5201:01:01:00",
                "activity_id": "CMIP",
                "branch_method": "Spin-up documentation",
                "branch_time_in_child": 1223115.0,
                "branch_time_in_parent": 1223115.0,
                "contact": "ec.cccma.info-info.ccmac.ec@canada.ca",
                "creation_date": "2019-04-30T17:18:16Z",
                "data_specs_version": "01.00.29",
                "experiment": "pre-industrial control",
                "experiment_id": "piControl",
                "external_variables": "areacella",
                "forcing_index": 1,
                "frequency": "mon",
                "further_info_url": "https://furtherinfo.es-doc.org/CMIP6.CCCma.CanESM5.piControl.none.r1i1p1f1",
                "grid_label": "gn",
                "initialization_index": 1,
                "institution_id": "CCCma",
                "mip_era": "CMIP6",
                "nominal_resolution": "500 km",
                "parent_activity_id": "CMIP",
                "parent_experiment_id": "piControl-spinup",
                "parent_mip_era": "CMIP6",
                "parent_source_id": "CanESM5",
                "parent_time_units": "days since 1850-01-01 0:0:0.0",
                "parent_variant_label": "r1i1p1f1",
                "physics_index": 1,
                "product": "model-output",
                "realization_index": 1,
                "realm": "atmos",
                "source_id": "CanESM5",
                "source_type": "AOGCM",
                "sub_experiment": "none",
                "sub_experiment_id": "none",
                "table_id": "Amon",
                "title": "CanESM5 output prepared for CMIP6",
                "tracking_id": "hdl:21.14100/cd5624dc-52ce-494b-9a12-bd3e6baaf468",
                "variable_id": "tas",
                "variant_label": "r1i1p1f1",
                "version": "v20190429",
                "cmor_version": "3.4.0",
            },
            "dims": {"time": 1, "lat": 1, "lon": 1, "bnds": 2},
            "data_vars": {
                "tas": {
                    "dims": ("time", "lat", "lon"),
                    "attrs": {
                        "standard_name": "air_temperature",
                        "long_name": "Near-Surface Air Temperature",
                        "units": "K",
                        "original_name": "ST",
                        "cell_methods": "area: time: mean",
                        "cell_measures": "area: areacella",
                    },
                    "data": [[[247.7601776123047]]],
                },
                "lat_bnds": {"dims": ("lat", "bnds"), "attrs": {}, "data": [[-90.0, -85.0]]},
                "lon_bnds": {"dims": ("lon", "bnds"), "attrs": {}, "data": [[-5.0, 5.0]]},
                "time_bnds": {"dims": ("time", "bnds"), "attrs": {}, "data": [[1223115.0, 1223146.0]]},
            },
        }
    ).to_netcdf(tmp_path / "tas_Amon_CanESM5_piControl_r1i1p1f1_gn_520101-534012.nc")

    child, parent = get_child_and_parent_dataset(
        df,
        parent_experiment="piControl",
        child_duration_in_years=140,
        parent_offset_in_years=0,
        parent_duration_in_years=140,
    )

    assert child == {
        "project": "CMIP6",
        "activity": "CMIP",
        "dataset": "CanESM5",
        "ensemble": "r1i1p1f1",
        "institute": "CCCma",
        "exp": "1pctCO2",
        "grid": "gn",
        "mip": "Amon",
        "timerange": "1850/1989",
    }
    assert parent == {
        "project": "CMIP6",
        "activity": "CMIP",
        "dataset": "CanESM5",
        "ensemble": "r1i1p1f1",
        "institute": "CCCma",
        "exp": "piControl",
        "grid": "gn",
        "mip": "Amon",
        "timerange": "5201/5340",
    }


@pytest.mark.parametrize(
    ("datasets", "expected"),
    [
        (
            pd.DataFrame(
                {
                    "instance_id": [
                        "CMIP6.ScenarioMIP.CCCma.CanESM5.ssp126.r1i1p1f1.Amon.pr.gn.v20190429",
                    ],
                    "source_id": ["CanESM5"],
                    "path": [
                        "pr_Amon_CanESM5_ssp126_r1i1p1f1_gn_210101-230012.nc",
                    ],
                }
            ),
            [
                "CMIP6/ScenarioMIP/CCCma/CanESM5/ssp126/r1i1p1f1/Amon/pr/gn/v20190429/pr_Amon_CanESM5_ssp126_r1i1p1f1_gn_210101-230012.nc",
            ],
        ),
        (
            pd.DataFrame(
                {
                    "instance_id": [
                        "obs4MIPs.obs4MIPs.ECMWF.ERA-5.ta.gn.v20250220",
                        "obs4MIPs.obs4MIPs.ECMWF.ERA-5.ta.gn.v20250220",
                    ],
                    "source_id": ["ERA-5", "ERA-5"],
                    "path": [
                        "ta_mon_ERA-5_PCMDI_gn_200701-200712.nc",
                        "ta_mon_ERA-5_PCMDI_gn_200801-200812.nc",
                    ],
                }
            ),
            [
                "obs4MIPs/ERA-5/v20250220/ta_mon_ERA-5_PCMDI_gn_200701-200712.nc",
                "obs4MIPs/ERA-5/v20250220/ta_mon_ERA-5_PCMDI_gn_200801-200812.nc",
            ],
        ),
    ],
)
def test_prepare_climate_data(tmp_path, datasets, expected):
    climate_data_dir = tmp_path / "climate_data"
    climate_data_dir.mkdir()

    source_data_dir = tmp_path / "source_data"
    source_data_dir.mkdir()

    datasets["path"] = [f"{source_data_dir / path}" for path in datasets["path"]]
    expected = [f"{climate_data_dir / path}" for path in expected]
    for path in datasets["path"]:
        Path(path).touch()

    prepare_climate_data(datasets, climate_data_dir)

    for source_path, symlink in zip(datasets["path"], expected):
        assert Path(symlink).is_symlink()
        assert Path(symlink).resolve() == Path(source_path).resolve()
